"""Custom helper class and functions to process CLI arguments with argparse."""

from __future__ import annotations
import argparse
import enum
from pathlib import Path
from typing import Any

from runrunner.base import Runner

from sparkle.platform.settings_objects import SettingState, Settings


class SetByUser(argparse.Action):
    """Possible action to execute for CLI argument."""

    def __call__(self: SetByUser, parser: argparse.ArgumentParser,
                 namespace: argparse.Namespace, values: str, option_string: str = None)\
            -> None:
        """Set attributes when called."""
        setattr(namespace, self.dest, values)
        setattr(namespace, self.dest + "_nondefault", True)


# Taken from https://stackoverflow.com/a/60750535
class EnumAction(argparse.Action):
    """Argparse action for handling Enums."""
    def __init__(self: EnumAction, **kwargs: str) -> None:
        """Initialise the EnumAction."""
        # Pop off the type value
        enum_type = kwargs.pop("type", None)

        # Ensure an Enum subclass is provided
        if enum_type is None:
            raise ValueError("type must be assigned an Enum when using EnumAction")
        if not issubclass(enum_type, enum.Enum):
            raise TypeError("type must be an Enum when using EnumAction")

        # Generate choices from the Enum
        kwargs.setdefault("choices", tuple(e.value for e in enum_type))

        super(EnumAction, self).__init__(**kwargs)

        self._enum = enum_type

    def __call__(self: EnumAction, parser: argparse.ArgumentParser,
                 namespace: argparse.Namespace, values: str, option_string: str = None) \
            -> None:
        """Converts value back to Enum."""
        value = self._enum(values)
        setattr(namespace, self.dest, value)


def user_set_state(args: argparse.Namespace, arg_name: str) -> SettingState:
    """Return the SettingState of an argument."""
    if hasattr(args, arg_name + "_nondefault"):
        return SettingState.CMD_LINE
    else:
        return SettingState.DEFAULT


def set_by_user(args: argparse.Namespace, arg_name: str) -> bool:
    """Return whether an argument was set through the CLI by the user or not."""
    return hasattr(args, arg_name + "_nondefault")


class ArgumentContainer():
    """Helper class for more convenient argument packaging and access."""
    def __init__(self: ArgumentContainer, names: list[str], kwargs: dict[str, Any])\
            -> None:
        """Create an ArgumentContainer.

        Args:
            names: List of names for the contained argument. For positional arguments,
                this will contain a single string. For positional arguments this will
                typically contain two, the first one starting with '-' and the second one
                starting with '--'.
            kwargs: Keyword arguments needed by the method parser.add_argument which adds
                the contained argument to a parser.
        """
        self.names = names
        self.kwargs = kwargs


AblationArgument = ArgumentContainer(names=["--ablation"],
                                     kwargs={"required": False,
                                             "action": "store_true",
                                             "help": "run ablation after configuration"})
SelectorAblationArgument =\
    ArgumentContainer(names=["--solver-ablation"],
                      kwargs={"required": False,
                              "action": "store_true",
                              "help": "construct a selector for "
                                      "each solver ablation combination"})

ActualMarginalContributionArgument = \
    ArgumentContainer(names=["--actual"],
                      kwargs={"action": "store_true",
                              "help": "compute the marginal contribution "
                                      "for the actual selector"})

AlsoConstructSelectorAndReportArgument = \
    ArgumentContainer(names=["--also-construct-selector-and-report"],
                      kwargs={"action": "store_true",
                              "help": "after running the solvers also construct the "
                                      "selector and generate the report"})

CleanupArgumentAll = \
    ArgumentContainer(names=["--all"],
                      kwargs={"action": "store_true",
                              "help": "clean all output files"})

CleanupArgumentRemove = \
    ArgumentContainer(names=["--remove"],
                      kwargs={"action": "store_true",
                              "help": "remove all files in the platform, including "
                                      "user data such as InstanceSets and Solvers"})

ConfiguratorArgument = ArgumentContainer(names=["--configurator"],
                                         kwargs={"type": Path,
                                                 "help": "path to configurator"})

CPUTimeArgument = \
    ArgumentContainer(names=["--cpu-time"],
                      kwargs={"type": int,
                              "help": "configuration budget per configurator run in "
                                      "seconds (cpu)"})

CutOffTimeArgument = \
    ArgumentContainer(names=["--cutoff-time"],
                      kwargs={"type": int,
                              "help": "The duration the portfolio will run before the "
                                      "solvers within the portfolio will be stopped "
                                      "(default: "
                                      f"{Settings.DEFAULT_general_target_cutoff_time})"})

DeterministicArgument =\
    ArgumentContainer(names=["--deterministic"],
                      kwargs={"action": "store_true",
                              "help": "Flag indicating the solver is deterministic"})

DownloadExamplesArgument =\
    ArgumentContainer(names=["--download-examples"],
                      kwargs={"action": argparse.BooleanOptionalAction,
                              "default": False,
                              "help": "Download the Examples into the directory."})

ExtractorPathArgument = ArgumentContainer(names=["extractor_path"],
                                          kwargs={"metavar": "extractor-path",
                                                  "type": str,
                                                  "help": "path or nickname of the "
                                                          "feature extractor"
                                                  })

GenerateJSONArgument = ArgumentContainer(names=["--only-json"],
                                         kwargs={"required": False,
                                                 "default": False,
                                                 "type": bool,
                                                 "help": "if set to True, only generate "
                                                         "machine readable output"
                                                 })

InstancePathPositional = ArgumentContainer(names=["instance_path"],
                                           kwargs={"type": Path,
                                                   "help": "Path to an instance (set)"})

InstancePath = ArgumentContainer(names=["--instance-path"],
                                 kwargs={"type": Path,
                                         "help": "Path to an instance (set)"})

InstanceSetTestArgument = \
    ArgumentContainer(names=["--instance-set-test"],
                      kwargs={"required": False,
                              "type": Path,
                              "help": "path to test instance set (only for validating)"})

InstanceSetTrainArgument = \
    ArgumentContainer(names=["--instance-set-train"],
                      kwargs={"required": True,
                              "type": Path,
                              "help": "path to training instance set"})

InstanceSetTestAblationArgument = \
    ArgumentContainer(names=["--instance-set-test"],
                      kwargs={"required": False,
                              "type": str,
                              "help": "path to test instance set"})

InstanceSetTrainAblationArgument = \
    ArgumentContainer(names=["--instance-set-train"],
                      kwargs={"required": False,
                              "type": str,
                              "help": "path to training instance set"})

InstanceSetTestReportArgument = \
    ArgumentContainer(names=["--instance-set-test"],
                      kwargs={"required": False,
                              "type": str,
                              "help": "path to testing instance set included in "
                              "Sparkle for an algorithm configuration report"})

InstanceSetTrainReportArgument = \
    ArgumentContainer(names=["--instance-set-train"],
                      kwargs={"required": False,
                              "type": str,
                              "help": "path to training instance set included in "
                              "Sparkle for an algorithm configuration report"})

InstancesPathArgument = ArgumentContainer(names=["instances_path"],
                                          kwargs={"metavar": "instances-path",
                                                  "type": str,
                                                  "help": "path to the instance set"})

InstancesPathRemoveArgument = \
    ArgumentContainer(names=["instances_path"],
                      kwargs={"metavar": "instances-path",
                              "type": str,
                              "help": "path to or nickname of the instance set"})

JobIDsArgument = ArgumentContainer(names=["--job-ids"],
                                   kwargs={"required": False,
                                           "nargs": "+",
                                           "type": str,
                                           "default": None,
                                           "help": "job ID to wait for"})

NicknameFeatureExtractorArgument = \
    ArgumentContainer(names=["--nickname"],
                      kwargs={"type": str,
                              "help": "set a nickname for the feature extractor"})

NicknameInstanceSetArgument = \
    ArgumentContainer(names=["--nickname"],
                      kwargs={"type": str,
                              "help": "set a nickname for the instance set"})

NicknamePortfolioArgument = \
    ArgumentContainer(names=["--portfolio-name"],
                      kwargs={"type": Path,
                              "help": "Specify a name of the portfolio. "
                                      "If none is given, one will be generated."})

NicknameSolverArgument = \
    ArgumentContainer(names=["--nickname"],
                      kwargs={"type": str,
                              "help": "set a nickname for the solver"})

NoAblationReportArgument = ArgumentContainer(names=["--no-ablation"],
                                             kwargs={"required": False,
                                                     "dest": "flag_ablation",
                                                     "default": True,
                                                     "const": False,
                                                     "nargs": "?",
                                                     "help": "turn off reporting on "
                                                             "ablation for an algorithm "
                                                             "configuration report"})

NumberOfRunsConfigurationArgument = \
    ArgumentContainer(names=["--number-of-runs"],
                      kwargs={"type": int,
                              "help": "number of configuration runs to execute"})

NumberOfRunsAblationArgument = \
    ArgumentContainer(names=["--number-of-runs"],
                      kwargs={"type": int,
                              "default": Settings.DEFAULT_config_number_of_runs,
                              "action": SetByUser,
                              "help": "Number of configuration runs to execute"})

PerfectSelectorMarginalContributionArgument =\
    ArgumentContainer(names=["--perfect"],
                      kwargs={"action": "store_true",
                              "help": "compute the marginal contribution "
                                      "for the perfect selector"})

RacingArgument = ArgumentContainer(names=["--racing"],
                                   kwargs={"type": bool,
                                           "default": Settings.
                                           DEFAULT_ablation_racing,
                                           "action": SetByUser,
                                           "help": "Performs abaltion analysis with "
                                                   "racing"})

RecomputeFeaturesArgument = \
    ArgumentContainer(names=["--recompute"],
                      kwargs={"action": "store_true",
                              "help": "Re-run feature extractor for instances with "
                                      "previously computed features"})

RecomputeMarginalContributionArgument = \
    ArgumentContainer(names=["--recompute"],
                      kwargs={"action": "store_true",
                              "help": "force marginal contribution to be recomputed even"
                                      " when it already exists in file for the current "
                                      "selector"})

RecomputeMarginalContributionForSelectorArgument = \
    ArgumentContainer(names=["--recompute-marginal-contribution"],
                      kwargs={"action": "store_true",
                              "help": "force marginal contribution to be recomputed even"
                                      " when it already exists in file for the current "
                                      "selector"})

RecomputePortfolioSelectorArgument = \
    ArgumentContainer(names=["--recompute-portfolio-selector"],
                      kwargs={"action": "store_true",
                              "help": "force the construction of a new portfolio "
                              "selector even when it already exists for the current "
                              "feature and performance data. NOTE: This will also "
                              "result in the computation of the marginal contributions "
                              "of solvers to the new portfolio selector."})

RecomputeRunSolversArgument = \
    ArgumentContainer(names=["--recompute"],
                      kwargs={"action": "store_true",
                              "help": "recompute the performance of all solvers on all "
                                      "instances"})

RunExtractorNowArgument = \
    ArgumentContainer(names=["--run-extractor-now"],
                      kwargs={"default": False,
                              "action": "store_true",
                              "help": "immediately run the feature extractor(s) on all "
                                      "the instances"})

RunOnArgument = ArgumentContainer(names=["--run-on"],
                                  kwargs={"type": Runner,
                                          "choices": [Runner.LOCAL,
                                                      Runner.SLURM],
                                          "action": EnumAction,
                                          "help": "On which computer or cluster "
                                                  "environment to execute the "
                                                  "calculation."})

RunSolverNowArgument = ArgumentContainer(names=["--run-solver-now"],
                                         kwargs={"default": False,
                                                 "action": "store_true",
                                                 "help": "immediately run the solver(s) "
                                                         "on all instances"})

SelectionReportArgument = \
    ArgumentContainer(names=["--selection"],
                      kwargs={"action": "store_true",
                              "help": "set to generate a normal selection report"})

SettingsFileArgument = \
    ArgumentContainer(names=["--settings-file"],
                      kwargs={"type": Path,
                              "default": Settings.DEFAULT_settings_path,
                              "action": SetByUser,
                              "help": "Specify the settings file to use in case you want"
                                      " to use one other than the default"})

SkipChecksArgument = ArgumentContainer(
    names=["--skip-checks"],
    kwargs={"dest": "run_checks",
            "default": True,
            "action": "store_false",
            "help": "Checks the solver's functionality by testing it on an instance "
                    "and the pcs file, when applicable."})

SnapshotArgument = ArgumentContainer(names=["snapshot_file_path"],
                                     kwargs={"metavar": "snapshot-file-path",
                                             "type": str,
                                             "help": "path to the snapshot file"})

SolverArgument = ArgumentContainer(names=["--solver"],
                                   kwargs={"required": True,
                                           "type": Path,
                                           "help": "path to solver"})

SolversArgument = ArgumentContainer(names=["--solvers"],
                                    kwargs={"required": False,
                                            "nargs": "+",
                                            "type": list[str],
                                            "help": "Specify the list of solvers to be "
                                                    "used. If not specifed, all solvers "
                                                    "known in Sparkle will be used."})

SolverCallsArgument = \
    ArgumentContainer(names=["--solver-calls"],
                      kwargs={"type": int,
                              "help": "number of solver calls to execute"})

SolverSeedsArgument = \
    ArgumentContainer(names=["--solver-seeds"],
                      kwargs={"type": int,
                              "help": "number of random seeds per solver to execute"})

SolverRemoveArgument = \
    ArgumentContainer(names=["solver"],
                      kwargs={"metavar": "solver",
                              "type": str,
                              "help": "name, path to or nickname of the solver"})

SolverPathArgument = ArgumentContainer(names=["solver_path"],
                                       kwargs={"metavar": "solver-path",
                                               "type": str,
                                               "help": "path to the solver"})

SolverReportArgument = ArgumentContainer(names=["--solver"],
                                         kwargs={"required": False,
                                                 "type": str,
                                                 "default": None,
                                                 "help": "path to solver for an "
                                                 "algorithm configuration report"})

TargetCutOffTimeAblationArgument = \
    ArgumentContainer(names=["--target-cutoff-time"],
                      kwargs={"type": int,
                              "default": Settings.DEFAULT_general_target_cutoff_time,
                              "action": SetByUser,
                              "help": "cutoff time per target algorithm run in seconds"})

TargetCutOffTimeConfigurationArgument = \
    ArgumentContainer(names=["--target-cutoff-time"],
                      kwargs={"type": int,
                              "help": "cutoff time per target algorithm run in seconds"})

TargetCutOffTimeRunSolversArgument = \
    ArgumentContainer(names=["--target-cutoff-time"],
                      kwargs={"type": int,
                              "help": "cutoff time per target algorithm run in seconds"})

TargetCutOffTimeValidationArgument = \
    ArgumentContainer(names=["--target-cutoff-time"],
                      kwargs={"type": int,
                              "default": Settings.DEFAULT_general_target_cutoff_time,
                              "action": SetByUser,
                              "help": "cutoff time per target algorithm run in seconds"})

TestCaseDirectoryArgument = \
    ArgumentContainer(names=["--test-case-directory"],
                      kwargs={"type": str,
                              "default": None,
                              "help": "Path to test case directory of an instance set "
                              + "for a selection report"})

UseFeaturesArgument = ArgumentContainer(names=["--use-features"],
                                        kwargs={"required": False,
                                                "action": "store_true",
                                                "help": "use the training set's features"
                                                        " for configuration"})

ValidateArgument = ArgumentContainer(names=["--validate"],
                                     kwargs={"required": False,
                                             "action": "store_true",
                                             "help": "validate after configuration"})

VerboseArgument = ArgumentContainer(names=["--verbose", "-v"],
                                    kwargs={"action": "store_true",
                                            "help": "output status in verbose mode"})

WallClockTimeArgument = \
    ArgumentContainer(names=["--wallclock-time"],
                      kwargs={"type": int,
                              "help": "configuration budget per configurator run in "
                                      "seconds (wallclock)"})

SelectorTimeoutArgument = \
    ArgumentContainer(names=["--selector-timeout"],
                      kwargs={"type": int,
                              "default": Settings.DEFAULT_portfolio_construction_timeout,
                              "help": "Cuttoff time (in seconds) for the algorithm"
                                      "selector construction"})

SparkleObjectiveArgument = \
    ArgumentContainer(names=["--objectives"],
                      kwargs={"type": str,
                              "help": "the comma seperated objective(s) to use."})



"""Class for solver status."""
from __future__ import annotations
from enum import Enum


class SolverStatus(str, Enum):
    """Possible return states for solver runs."""
    SUCCESS = "SUCCESS"
    CRASHED = "CRASHED"
    TIMEOUT = "TIMEOUT"
    WRONG = "WRONG"
    UNKNOWN = "UNKNOWN"
    ERROR = "ERROR"
    KILLED = "KILLED"

    # SAT specific status
    SAT = "SAT"
    UNSAT = "UNSAT"



"""This module provides tools for the argument parsing for solver wrappers."""
from pathlib import Path
import ast
from typing import Any

from sparkle.types import resolve_objective


def parse_commandline_dict(args: list[str]) -> dict:
    """Parses a commandline dictionary to the object."""
    dict_str = " ".join(args)
    dict_str = dict_str[dict_str.index("{"):dict_str.index("}") + 1]  # Slurm script fix
    return ast.literal_eval(dict_str)


def parse_solver_wrapper_args(args: list[str]) -> dict[Any]:
    """Parse the arguments passed to the solver wrapper.

    Args:
        args: a list of arguments passed via the command line. It is ensured by Sparkle
            that this list contains certain keys such as `solver_dir`.

    Returns:
        A dictionary mapping argument names to their currently held values.
    """
    args_dict = parse_commandline_dict(args)

    # Some data needs specific formatting
    args_dict["solver_dir"] = Path(args_dict["solver_dir"])
    args_dict["instance"] = Path(args_dict["instance"])
    args_dict["seed"] = int(args_dict["seed"])
    args_dict["objectives"] = [resolve_objective(name)
                               for name in args_dict["objectives"].split(",")]
    args_dict["cutoff_time"] = float(args_dict["cutoff_time"])

    if "config_path" in args_dict:
        # The arguments were not directly given and must be parsed from a file
        config_str = Path(args_dict["config_path"]).open("r")\
            .readlines()[args_dict["seed"]]
        # Extract the args without any quotes
        config_split = [arg.strip().replace("'", "").replace('"', "").strip("-")
                        for arg in config_str.split(" -") if arg.strip() != ""]
        for arg in config_split:
            varname, value = arg.strip("'").strip('"').split(" ", maxsplit=1)
            args_dict[varname] = value
        del args_dict["config_path"]

    return args_dict


def get_solver_call_params(args_dict: dict) -> list[str]:
    """Gather the additional parameters for the solver call.

    Args:
        args_dict: Dictionary mapping argument names to their currently held values

    Returns:
        A list of parameters for the solver call
    """
    params = []
    # Certain arguments are not relevant/have already been processed
    ignore_args = {"solver_dir", "instance", "cutoff_time", "seed", "objectives"}
    for key in args_dict:
        if key not in ignore_args and args_dict[key] is not None:
            params.extend(["-" + str(key), str(args_dict[key])])

    return params



"""This package provides specific Configurator class implementations for Sparkle."""
from sparkle.configurator.configurator import Configurator
from sparkle.configurator.implementations.smac2 import SMAC2, SMAC2Scenario


def resolve_configurator(configurator_name: str) -> Configurator:
    """Returns the Configurator subclass by name."""
    subclass_names = [SMAC2.__name__]
    if configurator_name in subclass_names:
        return eval(configurator_name)
    return None



#!/usr/bin/env python3
"""Sparkle CLI entry point."""
import sys
import os
from pathlib import Path


def commands() -> list[str]:
    """Get list of available commands."""
    module_path = Path(__file__).parent.resolve()
    self_name = Path(__file__).name
    return [path.stem for path in module_path.iterdir()
            if path.is_file() and path.suffix == ".py" and path.name != self_name]


def main() -> None:
    """Pass through command to launch CLI commands."""
    module_path = Path(__file__).parent.resolve()
    max_space = max([path.name.count("_") for path in module_path.iterdir()
                     if path.is_file()])
    if len(sys.argv) < 2:
        print("Usage: sparkle <command>")
        sys.exit(1)
    # Support spaces instead of _
    possible_commands = commands()
    for i in range(1, min(max_space, len(sys.argv))):
        command = "_".join(sys.argv[1:i + 1])
        args = sys.argv[i + 1:]
        command_file = module_path / f"{command}.py"
        if command in possible_commands:
            break
    if command_file.is_file():
        if not os.access(command_file, os.X_OK):  # Pip installation changes exec rights
            command_file.chmod(0o755)
        os.system(f"{command_file} {' '.join(args)}")
    else:
        print(f"Does not understand command {command}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python2.7
# encoding: utf-8

'''
generaicWrapper -- template for an AClib target algorithm wrapper

abstract methods for generation of callstring and parsing of solver output 

@author:     Marius Lindauer, Chris Fawcett, Alex Fréchette, Frank Hutter
        
@copyright:  2014 AClib. All rights reserved.
        
@license:    GPL

@contact:    lindauer@informatik.uni-freiburg.de, fawcettc@cs.ubc.ca, afrechet@cs.ubc.ca, fh@informatik.uni-freiburg.de

@note: example call: python src/generic_wrapper/spearWrapper.py --runsolver ./target_algorithms/runsolver/runsolver-3.3.4/src/runsolver -- <instance> <instance specific> <cutoff> <runlength> <seed>
@warning:  use "--" after the last additional argument of the wrapper to deactivate prefix matching! 
'''

import sys
import os
import signal
import time
import re
import random
import traceback
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from subprocess import Popen, PIPE
from tempfile import NamedTemporaryFile

__all__ = []
__version__ = 0.1
__authors__ = 'Marius Lindauer, Chris Fawcett, Alex Fréchette, Frank Hutter'
__date__ = '2014-03-06'
__updated__ = '2014-03-21'

def signalHandler(signum, frame):
    sys.exit(2)

class AbstractWrapper(object):
    '''
        abstract solver wrapper
    '''
    
    def __init__(self):
        '''
            Constructor
        '''
        #program_name = os.path.basename(sys.argv[0])
        program_version = "v%s" % __version__
        program_build_date = str(__updated__)
        program_version_message = "%%(prog)s %s (%s)" % (program_version, program_build_date)
        program_shortdesc = __import__("__main__").__doc__.split("\n")[1]
        program_license = '''%s
    
          Created by %s on %s.
          Copyright 2014 - AClib. All rights reserved.
          
          Licensed under the GPLv2
          http://www.gnu.org/licenses/gpl-2.0.html
          
          Distributed on an "AS IS" basis without warranties
          or conditions of any kind, either express or implied.
        
          USAGE
        ''' % (program_shortdesc, str(__authors__), str(__date__))
        self.parser = ArgumentParser(description=program_license, formatter_class=RawDescriptionHelpFormatter, add_help=False)
        self.args = None

        self.RESULT_MAPPING = {'SUCCESS': "SAT"} 
        self._watcher_file = None
        self._solver_file = None

        self._instance = ""
        self._specifics = ""
        self._cutoff = 0.0
        self._runlength = 0
        self._seed = 0
        
        self._exit_code = None
        
        self._runsolver = None
        self._mem_limit = 2048
        self._tmp_dir = None

        self._crashed_if_non_zero_status = True
        
        self._subprocesses = []
        
        self._DEBUG = True
        self._DELAY2KILL = 2

        self._ta_status = "EXTERNALKILL"
        self._ta_runtime = 999999999.0
        self._ta_runlength = -1
        self._ta_quality = -1
        self._ta_exit_code = None
        self._ta_misc = ""
        

    def print_d(self, str_):
        if self._DEBUG:
            print(str_)
        
    def main(self, argv=None): 
        ''' parse command line'''
        if argv is None:
            argv = sys.argv
        else:
            sys.argv.extend(argv)
    
        try:
            signal.signal(signal.SIGTERM, signalHandler)
            signal.signal(signal.SIGQUIT, signalHandler)
            signal.signal(signal.SIGINT, signalHandler)

            # Setup argument parser
            
            run_group = self.parser.add_argument_group("Run")
            run_group.add_argument("--runsolver-path", dest="runsolver", default=os.path.join(os.path.join(os.path.dirname(__file__),"runsolver"), "runsolver"), help="path to runsolver binary (if None, the runsolver is deactivated)")
            run_group.add_argument("--temp-file-dir", dest="tmp_dir", default=".", help="directory for temporary files (relative to -exec-dir in SMAC scenario)")
            run_group.add_argument("--mem-limit", dest="mem_limit", default=self._mem_limit, type=int, help="memory limit in MB")
            run_group.add_argument("--internal", dest="internal", default=False, action="store_true", help="skip calling an external target algorithm")
            
            run_group = self.parser.add_argument_group("External Callstring Generation and Output Parsing")
            run_group.add_argument("--ext-callstring", dest="ext_callstring", default=None, help="Command to get call string via external program;" +
                                                                                             "your programm gets a file with"+
                                                                                             "first line: instance name,"+
                                                                                             "second line: seed"+
                                                                                             "further lines: paramter name, paramater value;"+ 
                                                                                             "output: one line with callstring for target algorithm")
            run_group.add_argument("--ext-parsing", dest="ext_parsing", default=None, help="Command to use an external program to parse the output of your target algorihm;" +
                                                                                           "only paramter: name of output file;"+
                                                                                           "output of your progam:"+
                                                                                           "status: SAT|UNSAT|TIMEOUT|CRASHED\n"+
                                                                                           "quality: <integer>\n"+
                                                                                           "misc: <string>")

            help_group = self.parser.add_argument_group("Help")
            help_group.add_argument("--help", dest="show_help", action="store_true", help="show this help message")
            
            # Process arguments
            self.args, target_args = self.parser.parse_known_args()
            args = self.args
           
            if args.show_help:
                self.parser.print_help()
                self._ta_status = "ABORT"
                self._ta_misc = "help was requested..."
                self._exit_code = 1
                sys.exit(1)

            if args.runsolver != "None" and not os.path.isfile(args.runsolver) and not args.internal:
                self._ta_status = "ABORT"
                self._ta_misc = "runsolver is missing - should have been at %s." % (args.runsolver)
                self._exit_code = 1
                sys.exit(1)
            else:
                self._runsolver = args.runsolver
                self._mem_limit = args.mem_limit
            
            if not os.path.isdir(args.tmp_dir):
                self._ta_status = "ABORT"
                self._ta_misc = "temp directory is missing - should have been at %s." % (args.tmp_dir)
                self._exit_code = 1
                sys.exit(1)
            else:
                self._tmp_dir = args.tmp_dir
            
            if len(target_args) < 5:
                self._ta_status = "ABORT"
                self._ta_misc = "some required TA parameters (instance, specifics, cutoff, runlength, seed) missing - was [%s]." % (" ".join(target_args))
                self._exit_code = 1
                sys.exit(1)
                
            config_dict = self.build_parameter_dict(target_args)
            runargs = {
                        "instance": self._instance,
                        "specifics" : self._specifics,
                        "cutoff" : self._cutoff,
                        "runlength" : self._runlength,
                        "seed" : self._seed
                      }
            if args.ext_callstring:
                target_cmd = self.get_command_line_args_ext(runargs=runargs, config=config_dict, ext_call=args.ext_callstring).split(" ")
            else:
                target_cmd = self.get_command_line_args(runargs=runargs, config=config_dict).split(" ")
            
            if not args.internal:
                self.call_target(target_cmd)
                self.read_runsolver_output()
                
            if args.ext_parsing:
                resultMap = self.process_results_ext(self._solver_file, {"exit_code" : self._ta_exit_code}, ext_call=args.ext_parsing)
            else:
                resultMap = self.process_results(self._solver_file, {"exit_code" : self._ta_exit_code})
            
            if ('status' in resultMap):
                self._ta_status = self.RESULT_MAPPING.get(resultMap['status'],resultMap['status'])
            if ('runtime' in resultMap):
                self._ta_runtime = resultMap['runtime']
            if ('quality' in resultMap):
                self._ta_quality = resultMap['quality']
            if ('misc' in resultMap):
                self._ta_misc = resultMap['misc']
                
            # if still no status was determined, something went wrong and output files should be kept
            if self._ta_status is "EXTERNALKILL":
                self._ta_status = "CRASHED"
            sys.exit()
        except (KeyboardInterrupt, SystemExit):
            self.cleanup()
            self.print_result_string()
            if self._ta_exit_code:
                sys.exit(self._ta_exit_code)
            elif self._exit_code:
                sys.exit(self._exit_code)
            else:
                sys.exit(0)
        
    def build_parameter_dict(self, arg_list):
        '''
            Reads all arguments which were not parsed by ArgumentParser,
            extracts all meta information
            and builds a mapping: parameter name -> parameter value
            Format Assumption: <instance> <specifics> <runtime cutoff> <runlength> <seed> <solver parameters>
            Args:
                list of all options not parsed by ArgumentParser
        '''
        self._instance = arg_list[1]
        self._specifics = arg_list[2]
        self._cutoff = int(float(arg_list[3]) + 1) # runsolver only rounds down to integer
        self._runlength = int(arg_list[4])
        self._seed = int(arg_list[5])
        
        params = arg_list[6:]
        if (len(params)/2)*2 != len(params):
            self._ta_status = "ABORT"
            self._ta_misc = "target algorithm parameter list MUST have even length - found %d arguments." % (len(params))
            self.print_d(" ".join(params))
            self._exit_code = 1
            sys.exit(1)
        
        return dict((name, value) for name, value in zip(params[::2], params[1::2]))
        
    def call_target(self, target_cmd):
        '''
            extends the target algorithm command line call with the runsolver
            and executes it
            Args:
                list of target cmd (from getCommandLineArgs)
        '''
        random_id = random.randint(0,1000000)
        self._watcher_file = NamedTemporaryFile(suffix=".log", prefix="watcher-%d-" %(random_id), dir=self._tmp_dir, delete=False)
        self._solver_file = NamedTemporaryFile(suffix=".log", prefix="solver-%d-" %(random_id), dir=self._tmp_dir, delete=False)
        
        runsolver_cmd = []
        if self._runsolver != "None":
            runsolver_cmd = [self._runsolver, "-M", self._mem_limit, "-C", self._cutoff,
                             "-w", self._watcher_file.name,
                             "-o", self._solver_file.name]
        
        runsolver_cmd.extend(target_cmd)
        #for debugging
        self.print_d("Calling runsolver. Command-line:")
        self.print_d(" ".join(map(str,runsolver_cmd)))

        # run
        try:
            if self._runsolver != "None":
                io = Popen(map(str, runsolver_cmd), shell=False, preexec_fn=os.setpgrp)
            else:
                io = Popen(map(str, runsolver_cmd), stdout=self._solver_file, shell=False, preexec_fn=os.setpgrp)
            self._subprocesses.append(io)
            io.wait()
            self._subprocesses.remove(io)
            if io.stdout:
                io.stdout.flush()
        except OSError:
            self._ta_status = "ABORT"
            self._ta_misc = "execution failed: %s"  % (" ".join(map(str,runsolver_cmd)))
            self._exit_code = 1 
            sys.exit(1)
            
        self._solver_file.seek(0)

    def float_regex(self):
        return '[+-]?\d+(?:\.\d+)?(?:[eE][+-]\d+)?'

    def read_runsolver_output(self):
        '''
            reads self._watcher_file, 
            extracts runtime
            and returns if memout or timeout found
        ''' 
        if self._runsolver == "None":
            self._ta_exit_code = 0
            return
        
        self.print_d("Reading runsolver output from %s" % (self._watcher_file.name))
        data = self._watcher_file.read()

        if (re.search('runsolver_max_cpu_time_exceeded', data) or re.search('Maximum CPU time exceeded', data)):
            self._ta_status = "TIMEOUT"

        if (re.search('runsolver_max_memory_limit_exceeded', data)):
            self._ta_status = "TIMEOUT"
            self._ta_misc = "memory limit was exceeded"
           
        cpu_pattern1 = re.compile('runsolver_cputime: (%s)' % (self.float_regex()))
        cpu_match1 = re.search(cpu_pattern1, data)
            
        cpu_pattern2 = re.compile('CPU time \\(s\\): (%s)' % (self.float_regex()))
        cpu_match2 = re.search(cpu_pattern2, data)

        if (cpu_match1):
            self._ta_runtime = float(cpu_match1.group(1))
        if (cpu_match2):
            self._ta_runtime = float(cpu_match2.group(1))

        exitcode_pattern = re.compile('Child status: ([0-9]+)')
        exitcode_match = re.search(exitcode_pattern, data)

        if (exitcode_match):
            self._ta_exit_code = int(exitcode_match.group(1))

    def print_result_string(self):
        sys.stdout.write("Result for ParamILS: %s, %s, %s, %s, %s" % (self._ta_status, str(self._ta_runtime), str(self._ta_runlength), str(self._ta_quality), str(self._seed)))
        if (len(self._ta_misc) > 0):
            sys.stdout.write(", %s" % (self._ta_misc))

        print('')
        
    def cleanup(self):
        '''
            cleanup if error occurred or external signal handled
        '''
        if (len(self._subprocesses) > 0):
            print("killing the target run!")
            try:
                for sub in self._subprocesses:
                    #sub.terminate()
                    Popen(["pkill","-TERM", "-P",str(sub.pid)])
                    self.print_d("Wait %d seconds ..." % (self._DELAY2KILL))
                    time.sleep(self._DELAY2KILL)
                    if sub.returncode is None: # still running
                        sub.kill()

                self.print_d("done... If anything in the subprocess tree fork'd a new process group, we may not have caught everything...")
                self._ta_misc = "forced to exit by signal or keyboard interrupt."
                self._ta_runtime = self._cutoff
            except (OSError, KeyboardInterrupt, SystemExit):
                self._ta_misc = "forced to exit by multiple signals/interrupts."
                self._ta_runtime = self._cutoff

        if (self._ta_status is "ABORT" or self._ta_status is "CRASHED"):
            if (len(self._ta_misc) == 0):
                self._ta_misc = 'Problem with run. Exit code was %d.' % (self._ta_exit_code)

            if (self._watcher_file and self._solver_file):
                self._ta_misc = self._ta_misc + '; Preserving runsolver output at %s - preserving target algorithm output at %s' % (self._watcher_file.name or "<none>", self._solver_file.name or "<none>")

        try:
            if (self._watcher_file):
                self._watcher_file.close()
            if (self._solver_file):
                self._solver_file.close()

            if (self._ta_status is not "ABORT" and self._ta_status is not "CRASHED"):
                os.remove(self._watcher_file.name)
                os.remove(self._solver_file.name)
        except (OSError, KeyboardInterrupt, SystemExit):
            self._ta_misc = "problems removing temporary files during cleanup."
        except AttributeError:
            pass #in internal mode, these files are not generated
    
        if self._ta_status is "EXTERNALKILL":
            self._ta_status = "CRASHED"
            self._exit_code = 3

    def get_command_line_args(self, runargs, config):
        '''
        Returns the command call list containing arguments to execute the implementing subclass' solver.
        The default implementation delegates to get_command_line_args_ext. If this is not implemented, a
        NotImplementedError will be raised.
    
        Args:
            runargs: a map of any non-configuration arguments required for the execution of the solver.
            config: a mapping from parameter name (with prefix) to parameter value.
        Returns:
            A command call list to execute a target algorithm.
        '''
        raise NotImplementedError()

    def get_command_line_args_ext(self, runargs, config, ext_call):
        '''
        When production of the target algorithm is done from a source other than python,
        override this method to return a command call list to execute whatever you need to produce the command line.

        Args:
            runargs: a map of any non-configuration arguments required for the execution of the solver.
            config: a mapping from parameter name (with prefix) to parameter value.
            ext_call: string to call external program to get callstring of target algorithm
        Returns:
            A command call list to execute the command producing a single line of output containing the solver command string
        '''
        callstring_in = NamedTemporaryFile(suffix=".csv", prefix="callstring", dir=self._tmp_dir, delete=False)
        callstring_in.write("%s\n" %(runargs["instance"]))
        callstring_in.write("%d\n" %(runargs["seed"]))
        for name,value in config.items():
            callstring_in.write("%s,%s\n" %(name,value))
        callstring_in.flush()
        
        cmd = ext_call.split(" ")
        cmd.append(callstring_in.name)
        self.print_d(" ".join(cmd))
        try:
            io = Popen(cmd, shell=False, preexec_fn=os.setpgrp, stdout=PIPE)
            self._subprocesses.append(io)
            out_, _ = io.communicate()
            self._subprocesses.remove(io)
        except OSError:
            self._ta_misc = "failed to run external program for output parsing : %s" %(" ".join(cmd))
            self._ta_runtime = self._cutoff
            self._exit_code = 2
            sys.exit(2)
        if not out_ :
            self._ta_misc = "external program for output parsing yielded empty output: %s" %(" ".join(cmd))
            self._ta_runtime = self._cutoff
            self._exit_code = 2
            sys.exit(2)
        callstring_in.close()
        os.remove(callstring_in.name)
        return out_.strip("\n")
    
    def process_results(self, filepointer, out_args):
        '''
        Parse a results file to extract the run's status (SUCCESS/CRASHED/etc) and other optional results.
    
        Args:
            filepointer: a pointer to the file containing the solver execution standard out.
            exit_code : exit code of target algorithm
        Returns:
            A map containing the standard AClib run results. The current standard result map as of AClib 2.06 is:
            {
                "status" : <"SAT"/"UNSAT"/"TIMEOUT"/"CRASHED"/"ABORT">,
                "runtime" : <runtime of target algrithm>,
                "quality" : <a domain specific measure of the quality of the solution [optional]>,
                "misc" : <a (comma-less) string that will be associated with the run [optional]>
            }
            ATTENTION: The return values will overwrite the measured results of the runsolver (if runsolver was used). 
        '''
        raise NotImplementedError()

    def process_results_ext(self, filepointer, out_args, ext_call):
        '''
        Args:
            filepointer: a pointer to the file containing the solver execution standard out.
            exit_code : exit code of target algorithm
        Returns:
            A map containing the standard AClib run results. The current standard result map as of AClib 2.06 is:
            {
                "status" : <"SAT"/"UNSAT"/"TIMEOUT"/"CRASHED"/"ABORT">,
                "quality" : <a domain specific measure of the quality of the solution [optional]>,
                "misc" : <a (comma-less) string that will be associated with the run [optional]>
            }
        '''
        
        cmd = ext_call.split(" ")
        cmd.append(filepointer.name)
        self.print_d(" ".join(cmd))
        try:
            io = Popen(cmd, shell=False, preexec_fn=os.setpgrp, stdout=PIPE)
            self._subprocesses.append(io)
            out_, _ = io.communicate()
            self._subprocesses.remove(io)
        except OSError:
            self._ta_misc = "failed to run external program for output parsing"
            self._ta_runtime = self._cutoff
            self._exit_code = 2
            sys.exit(2)
        
        result_map = {}
        for line in out_.split("\n"):
            if line.startswith("status:"):
                result_map["status"] = line.split(":")[1].strip(" ")
            elif line.startswith("quality:"):
                result_map["quality"] = line.split(":")[1].strip(" ")
            elif line.startswith("misc:"):
                result_map["misc"] = line.split(":")[1]
        
        return result_map
        
#===============================================================================
# if __name__ == "__main__":
#     sys.exit(main())
#===============================================================================



#!/usr/bin/env python

import logging
import sys
import os
import inspect
cmd_folder = os.path.realpath(os.path.abspath(os.path.split(inspect.getfile( inspect.currentframe() ))[0]))
cmd_folder = os.path.realpath(os.path.join(cmd_folder, ".."))
if cmd_folder not in sys.path:
    sys.path.insert(0,cmd_folder)

import pickle
import argparse
from graphviz import Digraph
import traceback

def load_model(model_fn: str):
    '''
        loads saved model

        Arguments
        --------
        model_fn: str
            file name of saved model
            
        Returns
        -------
        scenario, feature_pre_pipeline, pre_solver, selector, config
    '''
    with open(model_fn, "br") as fp:
        scenario, feature_pre_pipeline, pre_solver, selector, config = pickle.load(
            fp)

    for fpp in feature_pre_pipeline:
        fpp.logger = logging.getLogger("Feature Preprocessing")
    if pre_solver:
        pre_solver.logger = logging.getLogger("Aspeed PreSolving")
    selector.logger = logging.getLogger("Selector")
    
    return scenario, feature_pre_pipeline, pre_solver, selector, config

def visualize(feature_pre_pipeline, pre_solver, selector):
    '''
        visualize all loaded components
        
        Arguments
        ---------
        feature_pre_pipeline: list
                list of fitted feature preprocessors
        pre_solver: Aspeed
                pre solver object with a saved static schedule
        selector: autofolio.selector.*
                fitted selector object
    '''
    
    dot = Digraph(comment='AutoFolio')
    fpp_idx = 0
    for fpp in feature_pre_pipeline:
        if not fpp.active:
            continue
        fpp_idx += 1
        dot.node('fpp_%d' %(fpp_idx), fpp.__class__.__name__)
        if fpp_idx > 0:
            dot.edge('fpp_%d' %(fpp_idx-1),'fpp_%d' %(fpp_idx))
        try:
            attributes = fpp.get_attributes()
            add_attributes(attributes=attributes, node_name='fpp_%d' %(fpp_idx), dot=dot)
        except AttributeError:
            #traceback.print_exc()
            pass
            
    for idx,presolver in enumerate(pre_solver.schedule):
        dot.node('pre_%d' %(idx), "%s for %d sec" %(presolver[0], presolver[1]))
        if idx > 0:
            dot.edge('pre_%d' %(idx-1),'pre_%d' %(idx))
        elif feature_pre_pipeline:
            dot.edge('fpp_%d' %(fpp_idx),'pre_%d' %(idx))
            
    dot.node("selector", selector.__class__.__name__)
    if pre_solver.schedule:
        dot.edge('pre_%d' %(len(pre_solver.schedule)-1), "selector")
    elif feature_pre_pipeline:
        dot.edge('fpp_%d' %(fpp_idx),'selector')
    try:
        attributes = selector.get_attributes()
        add_attributes(attributes=attributes, node_name='selector', dot=dot)
    except AttributeError:
        traceback.print_exc()
        pass
                   
        
    dot.render('test-output/autofolio', view=True)
    
def add_attributes(attributes, node_name:str, dot:Digraph):
    '''
        add attributes of <obj> with <node_name> to <dot>
        
        Arguments
        ---------
        attributes: str|list|dict
            attributes
        node_name: str
            node name of obj in dot
        dot: Digraph
            digraph of graphviz
    '''
    

    if isinstance(attributes,str):
        dot.node(attributes,attributes, shape='box', style='filled', color='lightgrey')
        dot.edge(node_name,attributes)
        return
    elif isinstance(attributes,list):
        for attr in attributes:
            add_attributes(attributes=attr, node_name=node_name, dot=dot)
    elif isinstance(attributes,dict):
        for k, v in attributes.items():
            dot.node(k,k, shape='box', style='filled', color='lightgrey')
            dot.edge(node_name,k)
            add_attributes(attributes=v, node_name=k, dot=dot)
    else:
        print("UNKNOWN TYPE: %s" %(attributes))
    
    
    #===========================================================================
    # for name, attr in attributes:
    #     if isinstance(attr, str):
    #         dot.node(name, "%s:%s"%(name, attr))
    #     elif isinstance(attr, list):
    #         dot.node(name, "%s"%(name))
    #         for s in attr:
    #             dot.node(str(s), str(s))
    #             dot.edge("%s"%(name), str(s))
    #     else:
    #         print("UNKNOWN TYPE: %s" %(attr))
    #         
    #     dot.edge(node_name,name)
    #===========================================================================
        
    

parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument("--load", type=str, default=None,
                         help="loads model (from --save); other modes are disabled with this options")
args = parser.parse_args()
        
scenario, feature_pre_pipeline, pre_solver, selector, config = load_model(args.load)

print(config)

visualize(feature_pre_pipeline, pre_solver, selector)






"""Test ablation scenario class."""


def test_ablation_scenario_constructor() -> None:
    """Test for ablation scenario constructor."""
    # TODO: Write test
    pass


def test_create_configuration_file() -> None:
    """Test for method create_configuration_file."""
    # TODO: Write test, check that written contents match with expected value
    pass


def test_create_instance_file() -> None:
    """Test for method create_instance_file."""
    # TODO: Write test for with training set
    # TODO: Write test for with test set
    pass


def test_check_for_ablation() -> None:
    """Test for method check_for_ablation."""
    # TODO: Write test for when correct file exists
    # TODO: Write test for when file does not exist
    # TODO: Write test for when file is corrupted
    pass


def test_read_ablation_table() -> None:
    """Test for reading an ablation table from file."""
    # TODO: Write test for correct file

    # TODO: Write test for not existant file

    # TODO: Write test for partially corrupted file
    pass


def test_submit_ablation() -> None:
    """Test for method submit ablation."""
    # NOTE: RunRunner calls must be mocked to avoid slurm job submission
    # TODO: Write test
    pass



#!/usr/bin/env python3
"""Sparkle command to construct a portfolio selector."""
import sys
import argparse
from pathlib import Path

import runrunner as rrr
from runrunner.base import Runner

from sparkle.CLI.help import global_variables as gv
from sparkle.structures import PerformanceDataFrame, FeatureDataFrame
from sparkle.CLI.help import logging as sl
from sparkle.platform.settings_objects import SettingState
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help.reporting_scenario import Scenario
from sparkle.platform import CommandName, COMMAND_DEPENDENCIES
from sparkle.CLI.initialise import check_for_initialise


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument(*ac.RecomputePortfolioSelectorArgument.names,
                        **ac.RecomputePortfolioSelectorArgument.kwargs)
    parser.add_argument(*ac.RecomputeMarginalContributionForSelectorArgument.names,
                        **ac.RecomputeMarginalContributionForSelectorArgument.kwargs)
    parser.add_argument(*ac.SelectorTimeoutArgument.names,
                        **ac.SelectorTimeoutArgument.kwargs)
    parser.add_argument(*ac.SparkleObjectiveArgument.names,
                        **ac.SparkleObjectiveArgument.kwargs)
    parser.add_argument(*ac.SelectorAblationArgument.names,
                        **ac.SelectorAblationArgument.kwargs)
    parser.add_argument(*ac.RunOnArgument.names,
                        **ac.RunOnArgument.kwargs)
    return parser


def judge_exist_remaining_jobs(feature_data_csv: Path,
                               performance_data_csv: Path) -> bool:
    """Return whether there are remaining feature or performance computation jobs."""
    feature_data = FeatureDataFrame(feature_data_csv)
    if feature_data.has_missing_vectors():
        return True
    performance_data = PerformanceDataFrame(performance_data_csv)
    return performance_data.has_missing_values


if __name__ == "__main__":
    # Log command call
    sl.log_command(sys.argv)

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()
    selector_timeout = args.selector_timeout
    flag_recompute_portfolio = args.recompute_portfolio_selector
    flag_recompute_marg_cont = args.recompute_marginal_contribution
    solver_ablation = args.solver_ablation

    check_for_initialise(
        COMMAND_DEPENDENCIES[CommandName.CONSTRUCT_PORTFOLIO_SELECTOR]
    )

    if ac.set_by_user(args, "objectives"):
        gv.settings().set_general_sparkle_objectives(
            args.objectives, SettingState.CMD_LINE
        )
    if args.run_on is not None:
        gv.settings().set_run_on(
            args.run_on.value, SettingState.CMD_LINE)
    run_on = gv.settings().get_run_on()

    print("Start constructing Sparkle portfolio selector ...")
    selector = gv.settings().get_general_sparkle_selector()

    flag_judge_exist_remaining_jobs = judge_exist_remaining_jobs(
        gv.settings().DEFAULT_feature_data_path,
        gv.settings().DEFAULT_performance_data_path)

    if flag_judge_exist_remaining_jobs:
        print("There remain unperformed feature computation jobs or performance "
              "computation jobs!")
        print("Please first execute all unperformed jobs before constructing Sparkle "
              "portfolio selector")
        print("Sparkle portfolio selector is not successfully constructed!")
        sys.exit(-1)

    # Selector (AutoFolio) cannot handle cutoff time less than 2, adjust if needed
    cutoff_time = max(gv.settings().get_general_target_cutoff_time(), 2)

    # Determine the objective function
    objective = gv.settings().get_general_sparkle_objectives()[0]

    performance_data = PerformanceDataFrame(gv.settings().DEFAULT_performance_data_path)
    feature_data = FeatureDataFrame(gv.settings().DEFAULT_feature_data_path)

    if feature_data.has_missing_value():
        print("WARNING: Missing values in the feature data, will be imputed as the mean "
              "value of all other non-missing values! Imputing all missing values...")
        feature_data.impute_missing_values()

    # TODO: Allow user to specify subsets of data to be used

    # Selector is named after the solvers it can predict, sort for permutation invariance
    solvers = sorted([s.name for s in gv.settings().DEFAULT_solver_dir.iterdir()])
    selection_scenario_path = (
        gv.settings().DEFAULT_selection_output
        / gv.settings().DEFAULT_general_sparkle_selector.name
        / "_".join(solvers))

    # Update latest scenario
    gv.latest_scenario().set_selection_scenario_path(selection_scenario_path)
    gv.latest_scenario().set_latest_scenario(Scenario.SELECTION)
    # Set to default to overwrite possible old path
    gv.latest_scenario().set_selection_test_case_directory()

    selector_path = selection_scenario_path / "portfolio_selector"
    sbatch_options = gv.settings().get_slurm_extra_options(as_args=True)
    if selector_path.exists() and not flag_recompute_portfolio:
        print("Portfolio selector already exists. Set the recompute flag to re-create.")
        sys.exit()

    selector_path.parent.mkdir(exist_ok=True, parents=True)

    selector_run = selector.construct(selector_path,
                                      performance_data,
                                      feature_data,
                                      objective,
                                      cutoff_time,
                                      selector_timeout,
                                      run_on=run_on,
                                      sbatch_options=sbatch_options,
                                      base_dir=sl.caller_log_dir)
    if run_on == Runner.LOCAL:
        print("Sparkle portfolio selector constructed!")
    else:
        print("Sparkle portfolio selector constructor running...")

    dependencies = [selector_run]
    if solver_ablation:
        for solver in performance_data.solvers:
            solver_name = Path(solver).name
            ablate_solver_dir = selection_scenario_path / f"ablate_{solver_name}"
            ablate_solver_selector = ablate_solver_dir / "portfolio_selector"
            if (ablate_solver_selector.exists() and not flag_recompute_portfolio):
                print(f"Portfolio selector without {solver_name} already exists. "
                      "Set the recompute flag to re-create.")
                continue
            ablate_solver_dir.mkdir(exist_ok=True, parents=True)
            ablated_performance_data = performance_data.copy()
            ablated_performance_data.remove_solver(solver)
            ablated_run = selector.construct(ablate_solver_selector,
                                             ablated_performance_data,
                                             feature_data,
                                             objective,
                                             cutoff_time,
                                             selector_timeout,
                                             run_on=run_on,
                                             sbatch_options=sbatch_options,
                                             base_dir=sl.caller_log_dir)
            dependencies.append(ablated_run)
            if run_on == Runner.LOCAL:
                print(f"Portfolio selector without {solver_name} constructed!")
            else:
                print(f"Portfolio selector without {solver_name} constructor running...")
    with_actual = "--actual" if solver_ablation else ""
    cmd = (f"sparkle/CLI/compute_marginal_contribution.py --perfect {with_actual} "
           f"{ac.SparkleObjectiveArgument.names[0]} {objective}")

    marginal_contribution = rrr.add_to_queue(
        runner=run_on,
        cmd=cmd,
        name=CommandName.COMPUTE_MARGINAL_CONTRIBUTION,
        base_dir=sl.caller_log_dir,
        dependencies=dependencies,
        sbatch_options=sbatch_options)
    dependencies.append(marginal_contribution)
    if run_on == Runner.LOCAL:
        marginal_contribution.wait()
        print("Selector marginal contribution computing done!")
    else:
        print(f"Running selector construction. Waiting for Slurm job(s) with id(s): "
              f"{', '.join([d.run_id for d in dependencies])}")

    # Write used settings to file
    gv.settings().write_used_settings()
    # Write used scenario to file
    gv.latest_scenario().write_scenario_ini()



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Configurator class to use different configurators like SMAC."""

from __future__ import annotations
from typing import Callable
from pathlib import Path
import ast
from statistics import mean
import operator
import fcntl
import glob
import shutil

import pandas as pd

import runrunner as rrr
from runrunner import Runner, Run

from sparkle.configurator.configurator import Configurator, ConfigurationScenario
from sparkle.solver import Solver
from sparkle.solver.validator import Validator
from sparkle.instance import InstanceSet
from sparkle.types import SparkleObjective


class SMAC2(Configurator):
    """Class for SMAC2 (Java) configurator."""
    configurator_path = Path(__file__).parent.parent.parent.resolve() /\
        "Components/smac-v2.10.03-master-778"
    target_algorithm = "smac_target_algorithm.py"

    def __init__(self: SMAC2,
                 objectives: list[SparkleObjective],
                 base_dir: Path,
                 output_path: Path) -> None:
        """Returns the SMAC configurator, Java SMAC V2.10.03.

        Args:
            objectives: The objectives to optimize. Only supports one objective.
            base_dir: The path where the configurator will be executed in.
            output_path: The path where the output will be placed.
        """
        output_path = output_path / SMAC2.__name__
        output_path.mkdir(parents=True, exist_ok=True)
        return super().__init__(
            validator=Validator(out_dir=output_path),
            output_path=output_path,
            executable_path=SMAC2.configurator_path / "smac",
            configurator_target=SMAC2.configurator_path / SMAC2.target_algorithm,
            objectives=objectives,
            base_dir=base_dir,
            tmp_path=output_path / "tmp",
            multi_objective_support=False)

    @property
    def scenario_class(self: Configurator) -> ConfigurationScenario:
        """Returns the SMAC2 scenario class."""
        return SMAC2Scenario

    def configure(self: Configurator,
                  scenario: ConfigurationScenario,
                  validate_after: bool = True,
                  sbatch_options: list[str] = [],
                  num_parallel_jobs: int = None,
                  base_dir: Path = None,
                  run_on: Runner = Runner.SLURM) -> list[Run]:
        """Start configuration job.

        Args:
            scenario: ConfigurationScenario object
            validate_after: Whether the Validator will be called after the configuration
            sbatch_options: List of slurm batch options to use
            num_parallel_jobs: The maximum number of jobs to run parallel.
            base_dir: The path where the sbatch scripts will be created for Slurm.
            run_on: On which platform to run the jobs. Default: Slurm.

        Returns:
            A RunRunner Run object.
        """
        self.scenario = scenario
        self.scenario.create_scenario(parent_directory=self.output_path)
        output_csv = self.scenario.validation / "configurations.csv"
        output_csv.parent.mkdir(exist_ok=True, parents=True)
        output = [f"{(self.scenario.result_directory).absolute()}/"
                  f"{self.scenario.name}_seed_{seed}_smac.txt"
                  for seed in range(self.scenario.number_of_runs)]
        cmds = [f"python3 {Configurator.configurator_cli_path.absolute()} "
                f"{SMAC2.__name__} {output[seed]} {output_csv.absolute()} "
                f"{self.executable_path.absolute()} "
                f"--scenario-file {(self.scenario.scenario_file_path).absolute()} "
                f"--seed {seed} "
                f"--execdir {self.scenario.tmp.absolute()}"
                for seed in range(self.scenario.number_of_runs)]
        parallel_jobs = self.scenario.number_of_runs
        if num_parallel_jobs is not None:
            parallel_jobs = max(num_parallel_jobs,
                                self.scenario.number_of_runs)
        configuration_run = rrr.add_to_queue(
            runner=run_on,
            cmd=cmds,
            name="configure_solver",
            base_dir=base_dir,
            output_path=output,
            parallel_jobs=parallel_jobs,
            sbatch_options=sbatch_options,
            srun_options=["-N1", "-n1"])
        runs = [configuration_run]

        if validate_after:
            self.validator.out_dir = output_csv.parent
            self.validator.tmp_out_dir = base_dir
            validate_run = self.validator.validate(
                [scenario.solver] * self.scenario.number_of_runs,
                output_csv.absolute(),
                [scenario.instance_set],
                [self.scenario.sparkle_objective],
                scenario.cutoff_time,
                subdir=Path(),
                dependency=configuration_run,
                sbatch_options=sbatch_options,
                run_on=run_on)
            runs.append(validate_run)

        if run_on == Runner.LOCAL:
            for run in runs:
                run.wait()
        return runs

    def get_optimal_configuration(
            self: Configurator,
            solver: Solver,
            instance_set: InstanceSet,
            objective: SparkleObjective = None,
            aggregate_config: Callable = mean) -> tuple[float, str]:
        """Returns optimal value and configuration string of solver on instance set."""
        if self.scenario is None:
            self.set_scenario_dirs(solver, instance_set)
        results = self.validator.get_validation_results(
            solver,
            instance_set,
            source_dir=self.scenario.validation,
            subdir=self.scenario.validation.relative_to(self.validator.out_dir))
        # Group the results per configuration
        if objective is None:
            objective = self.objectives[0]
        value_column = results[0].index(objective.name)
        config_column = results[0].index("Configuration")
        configurations = list(set(row[config_column] for row in results[1:]))
        config_scores = []
        for config in configurations:
            values = [float(row[value_column])
                      for row in results[1:] if row[1] == config]
            config_scores.append(aggregate_config(values))

        comparison = operator.lt if objective.minimise else operator.gt

        # Return optimal value
        min_index = 0
        current_optimal = config_scores[min_index]
        for i, score in enumerate(config_scores):
            if comparison(score, current_optimal):
                min_index, current_optimal = i, score

        # Return the optimal configuration dictionary as commandline args
        config_str = configurations[min_index].strip(" ")
        if config_str.startswith("{"):
            config = ast.literal_eval(config_str)
            config_str = " ".join([f"-{key} '{config[key]}'" for key in config])
        return current_optimal, config_str

    @staticmethod
    def organise_output(output_source: Path, output_target: Path = None) -> None | str:
        """Retrieves configurations from SMAC files and places them in output."""
        call_key = SMAC2.target_algorithm
        # Last line describing a call is the best found configuration
        for line in reversed(output_source.open("r").readlines()):
            if call_key in line:
                call_str = line.split(call_key, maxsplit=1)[1].strip()
                # The Configuration appears after the first 6 arguments
                configuration = call_str.split(" ", 7)[-1]
                if output_target is None:
                    return configuration
                with output_target.open("a") as fout:
                    fcntl.flock(fout.fileno(), fcntl.LOCK_EX)
                    fout.write(configuration + "\n")
                break

    def set_scenario_dirs(self: Configurator,
                          solver: Solver, instance_set: InstanceSet) -> None:
        """Patching method to allow the rebuilding of configuratio scenario."""
        self.scenario = self.scenario_class(solver, instance_set)
        self.scenario._set_paths(self.output_path)

    @staticmethod
    def get_smac_run_obj(objective: SparkleObjective) -> str:
        """Return the SMAC run objective based on the Performance Measure.

        Returns:
            A string that represents the run objective set in the settings.
        """
        if objective.time:
            return "RUNTIME"
        return "QUALITY"

    def get_status_from_logs(self: SMAC2) -> None:
        """Method to scan the log files of the configurator for warnings."""
        base_dir = self.output_path / "scenarios"
        if not base_dir.exists():
            return
        print(f"Checking the log files of configurator {type(self).__name__} for "
              "warnings...")
        scenarios = [f for f in base_dir.iterdir() if f.is_dir()]
        for scenario in scenarios:
            log_dir = scenario / "outdir_train_configuration" \
                / (scenario.name + "_scenario")
            warn_files = glob.glob(str(log_dir) + "/log-warn*")
            non_empty = [log_file for log_file in warn_files
                         if Path(log_file).stat().st_size > 0]
            if len(non_empty) > 0:
                print(f"Scenario {scenario.name} has {len(non_empty)} warning(s), see "
                      "the following log file(s) for more information:")
                for log_file in non_empty:
                    print(f"\t-{log_file}")
            else:
                print(f"Scenario {scenario.name} has no warnings.")


class SMAC2Scenario(ConfigurationScenario):
    """Class to handle SMAC2 configuration scenarios."""
    def __init__(self: ConfigurationScenario, solver: Solver,
                 instance_set: InstanceSet, number_of_runs: int = None,
                 solver_calls: int = None, cpu_time: int = None,
                 wallclock_time: int = None, cutoff_time: int = None,
                 cutoff_length: int = None,
                 sparkle_objectives: list[SparkleObjective] = None,
                 use_features: bool = None, configurator_target: Path = None,
                 feature_data_df: pd.DataFrame = None)\
            -> None:
        """Initialize scenario paths and names.

        Args:
            solver: Solver that should be configured.
            instance_set: Instances object for the scenario.
            number_of_runs: The number of configurator runs to perform
                for configuring the solver.
            solver_calls: The number of times the solver is called for each
                configuration run
            cpu_time: The time budget allocated for each configuration run. (cpu)
            wallclock_time: The time budget allocated for each configuration run.
                (wallclock)
            cutoff_time: The maximum time allowed for each individual run during
                configuration.
            cutoff_length: The maximum number of iterations allowed for each
                individual run during configuration.
            sparkle_objectives: SparkleObjectives used for each run of the configuration.
                Will be simplified to the first objective.
            use_features: Boolean indicating if features should be used.
            configurator_target: The target Python script to be called.
                This script standardises Configurator I/O for solver wrappers.
            feature_data_df: If features are used, this contains the feature data.
                Defaults to None.
        """
        super().__init__(solver, instance_set, sparkle_objectives)
        self.solver = solver
        self.instance_set = instance_set
        self.name = f"{self.solver.name}_{self.instance_set.name}"
        self.sparkle_objective = sparkle_objectives[0] if sparkle_objectives else None

        self.number_of_runs = number_of_runs
        self.solver_calls = solver_calls
        self.cpu_time = cpu_time
        self.wallclock_time = wallclock_time
        self.cutoff_time = cutoff_time
        self.cutoff_length = cutoff_length
        self.use_features = use_features
        self.configurator_target = configurator_target
        self.feature_data = feature_data_df

        self.parent_directory = Path()
        self.directory = Path()
        self.result_directory = Path()
        self.scenario_file_path = Path()
        self.feature_file_path = Path()
        self.instance_file_path = Path()

    def create_scenario(self: ConfigurationScenario, parent_directory: Path) -> None:
        """Create scenario with solver and instances in the parent directory.

        This prepares all the necessary subdirectories related to configuration.

        Args:
            parent_directory: Directory in which the scenario should be created.
        """
        self._set_paths(parent_directory)
        self._prepare_scenario_directory()
        self._prepare_result_directory()
        self._prepare_instances()

        if self.use_features:
            self._create_feature_file()

        self._create_scenario_file()

    def _set_paths(self: ConfigurationScenario, parent_directory: Path) -> None:
        """Set the paths for the scenario based on the specified parent directory."""
        self.parent_directory = parent_directory
        self.directory = self.parent_directory / "scenarios" / self.name
        self.result_directory = self.directory / "results"
        self.instance_file_path = self.directory / f"{self.instance_set.name}.txt"
        self.outdir_train = self.directory / "outdir_train_configuration"
        self.tmp = self.directory / "tmp"
        self.validation = self.directory / "validation"

    def _prepare_scenario_directory(self: ConfigurationScenario) -> None:
        """Delete old scenario dir, recreate it, create empty dirs inside."""
        shutil.rmtree(self.directory, ignore_errors=True)
        self.directory.mkdir(parents=True)

        # Create empty directories as needed
        self.outdir_train.mkdir()
        self.tmp.mkdir()

    def _prepare_result_directory(self: ConfigurationScenario) -> None:
        """Delete possible files in result directory."""
        shutil.rmtree(self.result_directory, ignore_errors=True)
        self.result_directory.mkdir(parents=True)

    def _create_scenario_file(self: ConfigurationScenario) -> None:
        """Create a file with the configuration scenario.

        Writes supplementary information to the target algorithm (algo =) as:
        algo = {configurator_target} {solver_directory} {sparkle_objective}
        """
        self.scenario_file_path = self.directory / f"{self.name}_scenario.txt"
        with self.scenario_file_path.open("w") as file:
            file.write(f"algo = {self.configurator_target.absolute()} "
                       f"{self.solver.directory.absolute()} {self.sparkle_objective} \n"
                       f"execdir = {self.tmp.absolute()}/\n"
                       f"deterministic = {1 if self.solver.deterministic else 0}\n"
                       f"run_obj = {self._get_performance_measure()}\n"
                       f"cutoffTime = {self.cutoff_time}\n"
                       f"cutoff_length = {self.cutoff_length}\n"
                       f"paramfile = {self.solver.get_pcs_file()}\n"
                       f"outdir = {self.outdir_train.absolute()}\n"
                       f"instance_file = {self.instance_file_path.absolute()}\n"
                       f"test_instance_file = {self.instance_file_path.absolute()}\n")
            if self.use_features:
                file.write(f"feature_file = {self.feature_file_path}\n")
            if self.wallclock_time is not None:
                file.write(f"wallclock-limit = {self.wallclock_time}\n")
            if self.cpu_time is not None:
                file.write(f"cputime-limit = {self.cpu_time}\n")
            if self.solver_calls is not None:
                file.write(f"runcount-limit = {self.solver_calls}\n")
            # We don't let SMAC do the validation
            file.write("validation = false" + "\n")

    def _prepare_instances(self: ConfigurationScenario) -> None:
        """Create instance list file without instance specifics."""
        self.instance_file_path.parent.mkdir(exist_ok=True, parents=True)
        with self.instance_file_path.open("w+") as file:
            for instance_path in self.instance_set._instance_paths:
                file.write(f"{instance_path.absolute()}\n")

    def _get_performance_measure(self: ConfigurationScenario) -> str:
        """Retrieve the performance measure of the SparkleObjective.

        Returns:
            Performance measure of the sparkle objective
        """
        if self.sparkle_objective.time:
            return "RUNTIME"
        return "QUALITY"

    def _create_feature_file(self: ConfigurationScenario) -> None:
        """Create CSV file from feature data."""
        self.feature_file_path = Path(self.directory
                                      / f"{self.instance_set.name}_features.csv")
        self.feature_data.to_csv(self.directory
                                 / self.feature_file_path, index_label="INSTANCE_NAME")

    def _clean_up_scenario_dirs(self: ConfigurationScenario,
                                configurator_path: Path,) -> list[Path]:
        """Yield directories to clean up after configuration scenario is done.

        Returns:
            list[str]: Full paths to directories that can be removed
        """
        result = []
        configurator_solver_path = configurator_path / "scenarios"\
            / f"{self.solver.name}_{self.instance_set.name}"

        for index in range(self.number_of_runs):
            dir = configurator_solver_path / str(index)
            result.append(dir)
        return result

    @staticmethod
    def from_file(scenario_file: Path, solver: Solver, instance_set: InstanceSet,
                  ) -> ConfigurationScenario:
        """Reads scenario file and initalises ConfigurationScenario."""
        config = {}
        with scenario_file.open() as file:
            for line in file:
                key, value = line.strip().split(" = ")
                config[key] = value

        # Collect relevant settings
        cpu_time = int(config["cpu_time"]) if "cpu_time" in config else None
        wallclock_limit = int(config["wallclock-limit"]) if "wallclock-limit" in config \
            else None
        solver_calls = int(config["runcount-limit"]) if "runcount-limit" in config \
            else None
        use_features = bool(config["feature_file"]) if "feature_file" in config \
            else None

        objective_str = config["algo"].split(" ")[-1]
        objective = SparkleObjective(objective_str)
        results_folder = scenario_file.parent / "results"
        state_run_dirs = [p for p in results_folder.iterdir() if p.is_file()]
        number_of_runs = len(state_run_dirs)
        return SMAC2Scenario(solver,
                             instance_set,
                             number_of_runs,
                             solver_calls,
                             cpu_time,
                             wallclock_limit,
                             int(config["cutoffTime"]),
                             config["cutoff_length"],
                             [objective],
                             use_features)



"""Init for the tools module."""
from sparkle.tools.general import get_time_pid_random_string
from sparkle.tools.pcsparser import PCSParser
from sparkle.tools.runsolver_parsing import get_solver_args
from sparkle.tools.slurm_parsing import SlurmBatch
from sparkle.tools.solver_wrapper_parsing import get_solver_call_params



"""Setup file for Sparkle."""
import os
from setuptools import setup, find_packages
from sparkle.about import version

setup(name="SparkleAI",
      version=version,
      url="https://github.com/thijssnelleman/Sparkle",
      author="Thijs Snelleman",
      author_email="fkt_sparkle@aim.rwth-aachen.de",
      keywords="ai sat planning",
      description="Sparkle is a Programming by Optimisation (PbO)-based problem-solving "
                  "platform designed to enable the widespread and effective use of PbO "
                  "techniques for improving the state-of-the-art in solving a broad "
                  "range of prominent AI problems, including SAT and AI Planning.",
      long_description=open("README.MD", "r").read() if os.path.exists("README.MD") else "",
      long_description_content_type="text/markdown",
      packages=find_packages(exclude=["tests"]),
      install_requires=[
          "numpy>=1.26.4",
          "pandas==2.2.2",
          "filelock==3.15.1",
          "tabulate==0.9.0",
          "tqdm==4.66.5",
          "RunRunner==0.1.6.4",
          # Reporting packages
          "plotly==5.23.0",
          "kaleido==0.2.1",
          # Autofolio packages
          "xgboost==2.0.3",
          "scikit-learn==1.3.2",
          "liac-arff==2.5.0",
          "smac==2.2.0"
      ],
      include_package_data=True,
      entry_points={"console_scripts": ["sparkle=sparkle.CLI.cli:main"], },)



"""Test files for Selector class."""


def test_selector_constructor() -> None:
    """Test for method constructor."""
    # TODO: Write test
    pass


def test_build_construction_cmd() -> None:
    """Test for method creation of the selector construction command."""
    # TODO: Write test, verify that output string is correct
    # Test with various runtime/wc time values and SparkleObjectives
    pass


def test_construct() -> None:
    """Test for method construct."""
    # NOTE: Make sure to mock RunRunner calls to avoid slurm job submission
    # TODO: Write test
    pass


def test_build_cmd() -> None:
    """Test for method build_cmd."""
    # TODO: Write test
    pass


def test_run() -> None:
    """Test for method run."""
    # NOTE: RunRunner calls must be mocked to avoid slurm job submission
    # TODO: Write test
    pass


def test_process_predict_schedule_output() -> None:
    """Test for method process_predict_schedule_output."""
    # TODO: Write test
    pass



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Helper functions for selection report generation."""
import sys
from pathlib import Path
from collections import Counter

from sparkle.CLI.compute_marginal_contribution\
    import compute_selector_marginal_contribution

from sparkle.platform import latex as stex
from sparkle.structures import PerformanceDataFrame, FeatureDataFrame
from sparkle.types.objective import SparkleObjective


def get_num_instance_sets(instance_list: list[str]) -> int:
    """Get the number of instance sets.

    Args:
        instance_list: List of instances to use

    Returns:
        The number of instance sets as LaTeX str.
    """
    return len(set([Path(instance_path).parent.name
                    for instance_path in instance_list]))


def get_instance_set_count_list(instance_list: list[str] = None) -> str:
    """Get the instance sets for use in a LaTeX document.

    Returns:
        The list of instance sets as LaTeX str.
    """
    instance_list = [Path(instance_path).parent.name for instance_path in instance_list]
    count = Counter(instance_list)
    rows = [(inst_key, f", constisting of {count[inst_key]} instances")
            for inst_key in count]
    return stex.list_to_latex(rows)


def solver_ranked_latex_list(solver_ranking: list[tuple[str, float]],
                             objective: SparkleObjective = None) -> str:
    """Convert a list of the solvers ranked by performance to LaTeX.

    Returns:
        The list of solvers ranked as LaTeX str.
    """
    objective_str = f"{objective}: " if objective is not None else ""
    return stex.list_to_latex([(row[0], f", {objective_str} {row[1]}")
                               for row in solver_ranking])


def get_portfolio_selector_performance(selection_scenario: Path) -> PerformanceDataFrame:
    """Creates a dictionary with the portfolio selector performance on each instance.

    Returns:
        A dict that maps instance name str to performance.
    """
    portfolio_selector_performance_path = selection_scenario / "performance.csv"
    if not portfolio_selector_performance_path.exists():
        print(f"ERROR: {portfolio_selector_performance_path} does not exist.")
        sys.exit(-1)
    return PerformanceDataFrame(portfolio_selector_performance_path)


def get_figure_portfolio_selector_vs_sbs(
        output_dir: Path,
        objective: SparkleObjective,
        train_data: PerformanceDataFrame,
        portfolio_selector_performance: PerformanceDataFrame,
        sbs_solver: str) -> str:
    """Create a LaTeX plot comparing the selector and the SBS.

    The plot compares the performance on each instance of the portfolio selector created
    by Sparkle and the SBS (single best solver).

    Returns:
        LaTeX str to include the comparison plot in a LaTeX report.
    """
    # We create a point of x,y form (SBS performance, portfolio performance)
    selector = portfolio_selector_performance.solvers[0]
    points = [[train_data.get_value(sbs_solver, instance, objective.name),
               portfolio_selector_performance.get_value(selector,
                                                        instance,
                                                        objective.name)]
              for instance in portfolio_selector_performance.instances]

    figure_filename = "figure_portfolio_selector_sparkle_vs_sbs"
    sbs_solver_name = Path(sbs_solver).name

    stex.generate_comparison_plot(points,
                                  figure_filename,
                                  xlabel=f"SBS ({sbs_solver_name}) [{objective}]",
                                  ylabel=f"Sparkle Selector [{objective}]",
                                  limit="magnitude",
                                  limit_min=0.25,
                                  limit_max=0.25,
                                  replace_zeros=True,
                                  output_dir=output_dir)
    return f"\\includegraphics[width=0.6\\textwidth]{{{figure_filename}}}"


def get_figure_portfolio_selector_sparkle_vs_vbs(
        output_dir: Path,
        objective: SparkleObjective,
        train_data: PerformanceDataFrame,
        actual_portfolio_selector_penalty: PerformanceDataFrame) -> str:
    """Create a LaTeX plot comparing the selector and the VBS.

    The plot compares the performance on each instance of the portfolio selector created
    by Sparkle and the VBS (virtual best solver).

    Returns:
        LaTeX str to include the comparison plot in a LaTeX report.
    """
    vbs_performance = train_data.best_instance_performance(objective=objective.name)
    instances = actual_portfolio_selector_penalty.instances
    solver = actual_portfolio_selector_penalty.solvers[0]
    points = [(vbs_performance[instance],
               actual_portfolio_selector_penalty.get_value(solver,
                                                           instance,
                                                           objective.name))
              for instance in instances]

    figure_filename = "figure_portfolio_selector_sparkle_vs_vbs"

    stex.generate_comparison_plot(points,
                                  figure_filename,
                                  xlabel=f"VBS [{objective}]",
                                  ylabel=f"Sparkle Selector [{objective.name}]",
                                  limit="magnitude",
                                  limit_min=0.25,
                                  limit_max=0.25,
                                  replace_zeros=True,
                                  output_dir=output_dir)
    return f"\\includegraphics[width=0.6\\textwidth]{{{figure_filename}}}"


def selection_report_variables(
        target_dir: Path,
        bibliograpghy_path: Path,
        extractor_path: Path,
        selection_scenario: Path,
        performance_data: PerformanceDataFrame,
        feature_data: FeatureDataFrame,
        objective: SparkleObjective,
        extractor_cutoff: int,
        cutoff: int,
        test_case_data: PerformanceDataFrame = None) -> dict[str, str]:
    """Returns: a dict matching variables in the LaTeX template with their values.

    Args:
        target_dir: Output path
        bibliography_path: Path to the bib file
        test_case_directory: Path to the test case directory.

    Returns:
        A dict matching str variables in the LaTeX template with their value str.
    """
    actual_performance_data = get_portfolio_selector_performance(selection_scenario)
    solver_performance_ranking = performance_data.get_solver_ranking(
        objective=objective)
    single_best_solver = solver_performance_ranking[0][0]
    latex_dict = {"bibliographypath": bibliograpghy_path.absolute(),
                  "numSolvers": performance_data.num_solvers,
                  "solverList": stex.list_to_latex([(s, "")
                                                    for s in performance_data.solvers])}
    latex_dict["numFeatureExtractors"] = len(
        [p for p in extractor_path.iterdir() if p.is_dir()])
    stex.list_to_latex([(f, "") for f in extractor_path.iterdir()])
    latex_dict["featureExtractorList"] = stex.list_to_latex(
        [(f, "") for f in extractor_path.iterdir()])
    latex_dict["numInstanceClasses"] = get_num_instance_sets(performance_data.instances)
    latex_dict["instanceClassList"] =\
        get_instance_set_count_list(performance_data.instances)
    latex_dict["featureComputationCutoffTime"] = extractor_cutoff
    latex_dict["performanceComputationCutoffTime"] = cutoff
    rank_list_perfect = performance_data.marginal_contribution(objective, sort=True)
    rank_list_actual = compute_selector_marginal_contribution(performance_data,
                                                              feature_data,
                                                              selection_scenario,
                                                              objective)
    latex_dict["solverPerfectRankingList"] = solver_ranked_latex_list(rank_list_perfect)
    latex_dict["solverActualRankingList"] = solver_ranked_latex_list(rank_list_actual)
    latex_dict["PARRankingList"] = solver_ranked_latex_list(solver_performance_ranking,
                                                            objective)
    latex_dict["VBSPAR"] = objective.instance_aggregator(
        performance_data.best_instance_performance(objective=objective.name))
    latex_dict["actualPAR"] = actual_performance_data.mean(objective=objective.name)
    latex_dict["metric"] = objective.name
    latex_dict["figure-portfolio-selector-sparkle-vs-sbs"] =\
        get_figure_portfolio_selector_vs_sbs(
            target_dir, objective, performance_data,
            actual_performance_data, single_best_solver)
    latex_dict["figure-portfolio-selector-sparkle-vs-vbs"] =\
        get_figure_portfolio_selector_sparkle_vs_vbs(target_dir,
                                                     objective,
                                                     performance_data,
                                                     actual_performance_data)
    latex_dict["testBool"] = r"\testfalse"

    # Train and test
    if test_case_data is not None:
        latex_dict["testInstanceClass"] =\
            f"\\textbf{ {test_case_data.csv_filepath.parent.name} }"
        latex_dict["numInstanceInTestInstanceClass"] =\
            test_case_data.num_instances
        latex_dict["testActualPAR"] = test_case_data.mean(objective=objective.name)
        latex_dict["testBool"] = r"\testtrue"

    return latex_dict


def generate_report_selection(target_path: Path,
                              latex_dir: Path,
                              latex_template: Path,
                              bibliography_path: Path,
                              extractor_path: Path,
                              selection_scenario: Path,
                              feature_data: FeatureDataFrame,
                              train_data: PerformanceDataFrame,
                              objective: SparkleObjective,
                              extractor_cutoff: int,
                              cutoff: int,
                              test_case_data: PerformanceDataFrame = None) -> None:
    """Generate a report for algorithm selection.

    Args:
        target_path: Path where the outputfiles will be placed.
        latex_dir: The latex dir
        latex_template: The template for the report
        bibliography_path: Path to the bib file.
        extractor_path: Path to the extractor used
        selection_scenario: Path to the selector scenario
        feature_data: Feature data created by extractor
        train_data: The performance input data for the selector
        objective: The objective for the selector
        extractor_cutoff: The maximum time for the selector to run
        cutoff: The cutoff per solver
        test_case_data: Path to the test case directory. Defaults to None.
    """
    # Include results on the test set if a test case directory is given
    latex_report_filename = Path("Sparkle_Report")
    if test_case_data is not None:
        latex_report_filename = Path("Sparkle_Report_for_Test")

    target_path.mkdir(parents=True, exist_ok=True)
    dict_variable_to_value = selection_report_variables(target_path,
                                                        bibliography_path,
                                                        extractor_path,
                                                        selection_scenario,
                                                        train_data,
                                                        feature_data,
                                                        objective,
                                                        extractor_cutoff,
                                                        cutoff,
                                                        test_case_data)
    stex.generate_report(latex_dir,
                         latex_template,
                         target_path,
                         latex_report_filename,
                         dict_variable_to_value)



#!/usr/bin/env python3
"""Sparkle command to validate a configured solver against its default configuration."""

import sys
import argparse
from pathlib import PurePath

from runrunner.base import Runner

from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.help import logging as sl
from sparkle.platform.settings_objects import Settings, SettingState
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help.reporting_scenario import Scenario
from sparkle.configurator.configurator import Configurator
from sparkle.solver.validator import Validator
from sparkle.solver import Solver
from sparkle.instance import instance_set
from sparkle.platform import CommandName, COMMAND_DEPENDENCIES
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help.nicknames import resolve_object_name


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(
        description=("Test the performance of the configured solver and the default "
                     "solver by doing validation experiments on the training and test "
                     "sets."))
    parser.add_argument(*ac.SolverArgument.names,
                        **ac.SolverArgument.kwargs)
    parser.add_argument(*ac.InstanceSetTrainArgument.names,
                        **ac.InstanceSetTrainArgument.kwargs)
    parser.add_argument(*ac.InstanceSetTestArgument.names,
                        **ac.InstanceSetTestArgument.kwargs)
    parser.add_argument(*ac.ConfiguratorArgument.names,
                        **ac.ConfiguratorArgument.kwargs)
    parser.add_argument(*ac.SparkleObjectiveArgument.names,
                        **ac.SparkleObjectiveArgument.kwargs)
    parser.add_argument(*ac.TargetCutOffTimeValidationArgument.names,
                        **ac.TargetCutOffTimeValidationArgument.kwargs)
    parser.add_argument(*ac.SettingsFileArgument.names,
                        **ac.SettingsFileArgument.kwargs)
    parser.add_argument(*ac.RunOnArgument.names,
                        **ac.RunOnArgument.kwargs)
    return parser


if __name__ == "__main__":
    # Log command call
    sl.log_command(sys.argv)

    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()
    solver = resolve_object_name(args.solver,
                                 gv.solver_nickname_mapping,
                                 gv.settings().DEFAULT_solver_dir,
                                 Solver)
    instance_set_train = resolve_object_name(
        args.instance_set_train,
        gv.file_storage_data_mapping[gv.instances_nickname_path],
        gv.settings().DEFAULT_instance_dir, instance_set)
    instance_set_test = resolve_object_name(
        args.instance_set_test,
        gv.file_storage_data_mapping[gv.instances_nickname_path],
        gv.settings().DEFAULT_instance_dir, instance_set)

    if args.run_on is not None:
        gv.settings().set_run_on(
            args.run_on.value, SettingState.CMD_LINE)
    run_on = gv.settings().get_run_on()

    check_for_initialise(
        COMMAND_DEPENDENCIES[CommandName.VALIDATE_CONFIGURED_VS_DEFAULT]
    )
    if args.configurator is not None:
        gv.settings().set_general_sparkle_configurator(
            value=getattr(Configurator, args.configurator),
            origin=SettingState.CMD_LINE)
    if ac.set_by_user(args, "settings_file"):
        gv.settings().read_settings_ini(
            args.settings_file, SettingState.CMD_LINE
        )  # Do first, so other command line options can override settings from the file

    if args.objectives is not None:
        gv.settings().set_general_sparkle_objectives(
            args.objectives, SettingState.CMD_LINE
        )
    if ac.set_by_user(args, "target_cutoff_time"):
        gv.settings().set_general_target_cutoff_time(
            args.target_cutoff_time, SettingState.CMD_LINE
        )
    # Compare current settings to latest.ini
    prev_settings = Settings(PurePath("Settings/latest.ini"))
    Settings.check_settings_changes(gv.settings(), prev_settings)

    # Make sure configuration results exist before trying to work with them
    configurator = gv.settings().get_general_sparkle_configurator()
    configurator.set_scenario_dirs(solver, instance_set_train)
    objectives = gv.settings().get_general_sparkle_objectives()
    # Record optimised configuration
    _, opt_config_str = configurator.get_optimal_configuration(
        solver, instance_set_train, objectives[0])
    opt_config = Solver.config_str_to_dict(opt_config_str)

    validator = Validator(gv.settings().DEFAULT_validation_output, sl.caller_log_dir)
    all_validation_instances = [instance_set_train]
    if instance_set_test is not None:
        all_validation_instances.append(instance_set_test)
    validation = validator.validate(
        solvers=[solver] * 2,
        configurations=[None, opt_config],
        instance_sets=all_validation_instances,
        objectives=objectives,
        cut_off=gv.settings().get_general_target_cutoff_time(),
        sbatch_options=gv.settings().get_slurm_extra_options(as_args=True),
        run_on=run_on)

    if run_on == Runner.LOCAL:
        validation.wait()
        print("Running validation done!")
    else:
        print(f"Running validation through Slurm with job ID: {validation.run_id}")

    # Update latest scenario
    gv.latest_scenario().set_config_solver(solver)
    gv.latest_scenario().set_config_instance_set_train(instance_set_train.directory)
    gv.latest_scenario().set_latest_scenario(Scenario.CONFIGURATION)

    if instance_set_test is not None:
        gv.latest_scenario().set_config_instance_set_test(instance_set_test.directory)
    else:
        # Set to default to overwrite possible old path
        gv.latest_scenario().set_config_instance_set_test()

    # Write used settings to file
    gv.settings().write_used_settings()
    # Write used scenario to file
    gv.latest_scenario().write_scenario_ini()



"""Test class for the InstanceSet class."""


def test_resolve_instance_set() -> None:
    """Test for resolving the correct instance set subclass."""
    # TODO: Write test for MultiFileInstanceSet
    # TODO: Write test for IterableFileInstanceSet
    # TODO: Write test for FileInstanceSet
    pass


def test_multi_file_instance_set() -> None:
    """Test for MultiFileInstanceSet properties."""
    # TODO: Write test
    pass


def test_iterable_file_instance_set() -> None:
    """Test for IterableFileInstanceSet properties."""
    # TODO: Write test
    pass


def test_file_instance_set() -> None:
    """Test for FileInstanceSet properties."""
    # TODO: Write test
    pass



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Helper functions for algorithm configuration report generation."""
from __future__ import annotations

import sys
from pathlib import Path

from sparkle.platform import latex as stex
from sparkle.solver.ablation import AblationScenario
from sparkle.solver.validator import Validator
from sparkle.configurator.configurator import Configurator, ConfigurationScenario
from sparkle.solver import Solver
from sparkle.instance import InstanceSet
from sparkle.configurator.implementations import SMAC2
from sparkle.types import SparkleObjective
from sparkle import about


def get_features_bool(configurator_scenario: ConfigurationScenario,
                      solver_name: str, train_set: InstanceSet) -> str:
    """Return a bool string for latex indicating whether features were used.

    True if a feature file is given in the scenario file, false otherwise.

    Args:
        solver_name: Name of the solver
        instance_set_train_name: Name of the instance set used for training

    Returns:
        A string describing whether features are used
    """
    scenario_file = configurator_scenario.directory \
        / f"{solver_name}_{train_set.name}_scenario.txt"

    for line in scenario_file.open("r").readlines():
        if line.split(" ")[0] == "feature_file":
            return "\\featurestrue"
    return "\\featuresfalse"


def get_average_performance(results: list[list[str]],
                            objective: SparkleObjective) -> float:
    """Return the PAR score for a given results file and cutoff time.

    Args:
        results_file: Name of the result file
        objective: The objective to average

    Returns:
        Average performance value
    """
    instance_per_dict = get_dict_instance_to_performance(results,
                                                         objective)
    num_instances = len(instance_per_dict.keys())
    sum_par = sum(float(instance_per_dict[instance]) for instance in instance_per_dict)
    return float(sum_par / num_instances)


def get_dict_instance_to_performance(results: list[list[str]],
                                     objective: SparkleObjective) -> dict[str, float]:
    """Return a dictionary of instance names and their performance.

    Args:
        results: Results from CSV
        objective: The Sparkle Objective we are converting for
    Returns:
        A dictionary containing the performance for each instance
    """
    value_column = results[0].index(objective.name)
    results_per_instance = {}
    for row in results[1:]:
        value = float(row[value_column])
        results_per_instance[Path(row[3]).name] = value
    return results_per_instance


def get_ablation_bool(scenario: AblationScenario) -> str:
    """Return the ablation bool as LaTeX string.

    Args:
        solver: The solver object
        instance_train_name: Name of the trianing instance set
        instance_test_name: Name of the testing instance set

    Returns:
        A string describing whether ablation was run or not
    """
    if scenario.check_for_ablation():
        return "\\ablationtrue"
    return "\\ablationfalse"


def get_data_for_plot(configured_results: list[list[str]],
                      default_results: list[list[str]],
                      objective: SparkleObjective) -> list:
    """Return the required data to plot.

    Creates a nested list of performance values algorithm runs with default and
    configured parameters on instances in a given instance set.

    Args:
        configured_results_dir: Directory of results for configured solver
        default_results_dir: Directory of results for default solver
        run_cutoff_time: Cutoff time

    Returns:
        A list of lists containing data points
    """
    dict_instance_to_par_default = get_dict_instance_to_performance(
        default_results, objective)
    dict_instance_to_par_configured = get_dict_instance_to_performance(
        configured_results, objective)

    instances = (dict_instance_to_par_default.keys()
                 & dict_instance_to_par_configured.keys())
    if (len(dict_instance_to_par_default) != len(instances)):
        print("""ERROR: Number of instances does not match
         the number of performance values for the default configuration.""")
        sys.exit(-1)
    points = []
    for instance in instances:
        point = [dict_instance_to_par_default[instance],
                 dict_instance_to_par_configured[instance]]
        points.append(point)

    return points


def get_figure_configure_vs_default(configured_results: list[list[str]],
                                    default_results: list[list[str]],
                                    target_directory: Path,
                                    figure_filename: str,
                                    performance_measure: str,
                                    run_cutoff_time: float,
                                    objective: SparkleObjective) -> str:
    """Create a figure comparing the configured and default solver.

    Base function to create a comparison plot of a given instance set between the default
    and configured performance.

    Args:
        configured_results_dir: Directory of results for configured solver
        default_results_dir: Directory of results for default solver
        target_directory: Directory for the configuration reports
        figure_filename: Filename for the figure
        run_cutoff_time: Cutoff time

    Returns:
        A string containing the latex command to include the figure
    """
    points = get_data_for_plot(configured_results, default_results,
                               objective)

    plot_params = {"xlabel": f"Default parameters [{performance_measure}]",
                   "ylabel": f"Configured parameters [{performance_measure}]",
                   "scale": "linear",
                   "limit_min": 1.5,
                   "limit_max": 1.5,
                   "limit": "relative",
                   "replace_zeros": False,
                   "output_dir": target_directory
                   }
    if performance_measure[0:3] == "PAR":
        plot_params["scale"] = "log"
        plot_params["limit_min"] = 0.25
        plot_params["limit_max"] = 0.25
        plot_params["limit"] = "magnitude"
        plot_params["replace_zeros"] = True

    stex.generate_comparison_plot(points,
                                  figure_filename,
                                  **plot_params)

    return f"\\includegraphics[width=0.6\\textwidth]{{{figure_filename}}}"


def get_figure_configured_vs_default_on_instance_set(solver: Solver,
                                                     instance_set_name: str,
                                                     res_default: list[list[str]],
                                                     res_conf: list[list[str]],
                                                     target_directory: Path,
                                                     smac_objective: str,
                                                     run_cutoff_time: float,
                                                     objective: SparkleObjective,
                                                     data_type: str = "train") -> str:
    """Create a figure comparing the configured and default solver on the training set.

    Manages the creation of a comparison plot of the instances in the train instance set
    for the report by gathering the proper files and choosing the plotting parameters
    based on the performance measure.

    Args:
        solver: The solver object
        instance_set_train_name: Name of the instance set for training
        configuration_reports_directory: Directory to the configuration reports
        run_cutoff_time: Cutoff time

    Returns:
        A string containing the latex comand to include the figure
    """
    data_plot_configured_vs_default_on_instance_set_filename = (
        f"data_{solver.name}_configured_vs_default_on_{instance_set_name}_{data_type}")
    return get_figure_configure_vs_default(
        res_conf, res_default, target_directory,
        data_plot_configured_vs_default_on_instance_set_filename,
        smac_objective,
        run_cutoff_time,
        objective)


def get_timeouts_instanceset(solver: Solver,
                             instance_set: InstanceSet,
                             configurator: Configurator,
                             validator: Validator,
                             cutoff: float) -> tuple[int, int, int]:
    """Return the number of timeouts by configured, default and both on the testing set.

    Args:
        solver: The solver object
        instance_set: Instance Set
        configurator: Configurator
        validator: Validator
        cutoff: Cutoff time

    Returns:
        A tuple containing the number of timeouts for the different configurations
    """
    objective = configurator.scenario.sparkle_objective
    _, config = configurator.get_optimal_configuration(
        solver, instance_set, objective)
    res_default = validator.get_validation_results(solver,
                                                   instance_set,
                                                   config="")
    res_conf = validator.get_validation_results(solver,
                                                instance_set,
                                                config=config)
    dict_instance_to_par_configured = get_dict_instance_to_performance(
        res_conf, objective)
    dict_instance_to_par_default = get_dict_instance_to_performance(
        res_default, objective)

    return get_timeouts(dict_instance_to_par_configured,
                        dict_instance_to_par_default, cutoff)


def get_timeouts(instance_to_par_configured: dict,
                 instance_to_par_default: dict,
                 cutoff: float) -> tuple[int, int, int]:
    """Return the number of timeouts for given dicts.

    Args:
        dict_instance_to_par_configured: _description_
        dict_instance_to_par_default: _description_
        cutoff: Cutoff value

    Returns:
        A tuple containing timeout values
    """
    configured_timeouts = 0
    default_timeouts = 0
    overlapping_timeouts = 0

    for instance in instance_to_par_configured:
        configured_par = instance_to_par_configured[instance]
        default_par = instance_to_par_default[instance]
        # Count the amount of values that are equal to timeout
        configured_timeouts += (configured_par > cutoff)
        default_timeouts += (default_par > cutoff)
        overlapping_timeouts += (configured_par > cutoff
                                 and default_par > cutoff)

    return configured_timeouts, default_timeouts, overlapping_timeouts


def get_ablation_table(scenario: AblationScenario) -> str:
    """Generate a LaTeX table of the ablation path.

    This is the result of the ablation analysis to determine the parameter importance.

    Args:
        solver: The solver object
        instance_set_train_name: Name of the instance set for training
        instance_set_test_name: Name of the instance set for testing

    Returns:
        A string containing the LaTeX table code of the ablation path
    """
    results = scenario.read_ablation_table()
    table_string = r"\begin{tabular}{rp{0.25\linewidth}rrr}"
    # "Round", "Flipped parameter", "Source value", "Target value", "Validation result"
    for i, line in enumerate(results):
        # If this fails something has changed in the representation of ablation tables
        if len(line) != 5:
            print("""ERROR: something has changed with the representation
                   of ablation tables""")
            sys.exit(-1)
        if i == 0:
            line = [f"\\textbf{{{word}}}" for word in line]

        # Put multiple variable changes in one round on a seperate line
        if (len(line[1].split(",")) > 1
           and len(line[1].split(",")) == len(line[2].split(","))
           and len(line[1].split(",")) == len(line[3].split(","))):
            params = line[1].split(",")
            default_values = line[2].split(",")
            flipped_values = line[3].split(",")

            sublines = len(params)
            for subline in range(sublines):
                round = "" if subline != 0 else line[0]
                result = "" if subline + 1 != sublines else line[-1]
                printline = [round, params[subline], default_values[subline],
                             flipped_values[subline], result]
                table_string += " & ".join(printline) + " \\\\ "
        else:
            table_string += " & ".join(line) + " \\\\ "
        if i == 0:
            table_string += "\\hline "
    table_string += "\\end{tabular}"

    return table_string


def configuration_report_variables(target_dir: Path,
                                   solver: Solver,
                                   configurator: Configurator,
                                   validator: Validator,
                                   extractor_dir: Path,
                                   bib_path: Path,
                                   instance_set_train: InstanceSet,
                                   extractor_cuttoff: int,
                                   instance_set_test: InstanceSet = None,
                                   ablation: AblationScenario = None) -> dict:
    """Return a dict matching LaTeX variables and their values.

    Args:
        solver: Object representation of the Solver
        instance_set_train: Path of the instance set for training
        instance_set_test: Path of the instance set for testing. Defaults to None.
        ablation: Whether or not ablation is used. Defaults to True.

    Returns:
        A dictionary containing the variables and values
    """
    has_test = instance_set_test is not None

    full_dict = get_dict_variable_to_value_common(solver,
                                                  configurator,
                                                  validator,
                                                  ablation,
                                                  bib_path,
                                                  instance_set_train,
                                                  target_dir)

    if has_test:
        test_dict = get_dict_variable_to_value_test(target_dir,
                                                    solver,
                                                    configurator,
                                                    validator,
                                                    ablation,
                                                    instance_set_train,
                                                    instance_set_test)
        full_dict.update(test_dict)
    full_dict["testBool"] = f"\\test{str(has_test).lower()}"

    if ablation is None:
        full_dict["ablationBool"] = "\\ablationfalse"

    if full_dict["featuresBool"] == "\\featurestrue":
        full_dict["numFeatureExtractors"] =\
            len([p for p in extractor_dir.iterdir()])
        full_dict["featureExtractorList"] =\
            stex.list_to_latex([(p.name, "") for p in extractor_dir.iterdir()])
        full_dict["featureComputationCutoffTime"] = extractor_cuttoff

    return full_dict


def get_dict_variable_to_value_common(solver: Solver,
                                      configurator: Configurator,
                                      validator: Validator,
                                      ablation: AblationScenario,
                                      bibliography_path: Path,
                                      train_set: InstanceSet,
                                      target_directory: Path) -> dict:
    """Return a dict matching LaTeX variables and values used for all config. reports.

    Args:
        Solver: The solver object
        instance_set_train: Path of the instance set for training
        instance_set_test: Path of the instance set for testing
        target_directory: Path to directory with configuration reports

    Returns:
        A dictionary containing the variables and values
    """
    objective = configurator.scenario.sparkle_objective
    _, opt_config = configurator.get_optimal_configuration(
        solver, train_set, objective)
    res_default = validator.get_validation_results(
        solver, train_set, config="")
    res_conf = validator.get_validation_results(
        solver, train_set, config=opt_config)
    instance_names = set([res[3] for res in res_default])

    latex_dict = {"bibliographypath": bibliography_path.absolute()}
    latex_dict["performanceMeasure"] = objective.name
    smac_run_obj = SMAC2.get_smac_run_obj(objective)

    if smac_run_obj == "RUNTIME":
        latex_dict["runtimeBool"] = "\\runtimetrue"
    elif smac_run_obj == "QUALITY":
        latex_dict["runtimeBool"] = "\\runtimefalse"

    latex_dict["solver"] = solver.name
    latex_dict["instanceSetTrain"] = train_set.name
    latex_dict["sparkleVersion"] = about.version
    latex_dict["numInstanceInTrainingInstanceSet"] = len(instance_names)

    run_cutoff_time = configurator.scenario.cutoff_time
    latex_dict["numSmacRuns"] = configurator.scenario.number_of_runs
    latex_dict["smacObjective"] = smac_run_obj
    latex_dict["smacWholeTimeBudget"] = configurator.scenario.wallclock_time
    latex_dict["smacEachRunCutoffTime"] = run_cutoff_time
    latex_dict["optimisedConfiguration"] = opt_config
    latex_dict["optimisedConfigurationTrainingPerformancePAR"] =\
        get_average_performance(res_conf, objective)
    latex_dict["defaultConfigurationTrainingPerformancePAR"] =\
        get_average_performance(res_default, objective)

    str_value = get_figure_configured_vs_default_on_instance_set(
        solver, train_set.name, res_default, res_conf, target_directory,
        smac_run_obj, float(run_cutoff_time), objective)
    latex_dict["figure-configured-vs-default-train"] = str_value

    # Retrieve timeout numbers for the training instances
    configured_timeouts_train, default_timeouts_train, overlapping_timeouts_train =\
        get_timeouts_instanceset(solver, train_set, configurator, validator,
                                 run_cutoff_time)

    latex_dict["timeoutsTrainDefault"] = default_timeouts_train
    latex_dict["timeoutsTrainConfigured"] = configured_timeouts_train
    latex_dict["timeoutsTrainOverlap"] = overlapping_timeouts_train
    latex_dict["ablationBool"] = get_ablation_bool(ablation)
    latex_dict["ablationPath"] = get_ablation_table(ablation)
    latex_dict["featuresBool"] = get_features_bool(
        configurator.scenario, solver.name, train_set)

    return latex_dict


def get_dict_variable_to_value_test(target_dir: Path,
                                    solver: Solver,
                                    configurator: Configurator,
                                    validator: Validator,
                                    ablation: AblationScenario,
                                    train_set: InstanceSet,
                                    test_set: InstanceSet) -> dict:
    """Return a dict matching test set specific latex variables with their values.

    Args:
        target_dir: Path to where output should go
        solver: The solver object
        configurator: Configurator for which the report is generated
        validator: Validator that provided the data set results
        train_set: Instance set for training
        test_set: Instance set for testing

    Returns:
        A dictionary containting the variables and their values
    """
    _, config = configurator.get_optimal_configuration(
        solver, train_set, configurator.scenario.sparkle_objective)
    res_default = validator.get_validation_results(
        solver, test_set, config="")
    res_conf = validator.get_validation_results(
        solver, test_set, config=config)
    instance_names = set([res[3] for res in res_default])
    run_cutoff_time = configurator.scenario.cutoff_time
    objective = configurator.scenario.sparkle_objective
    test_dict = {"instanceSetTest": test_set.name}
    test_dict["numInstanceInTestingInstanceSet"] = len(instance_names)
    test_dict["optimisedConfigurationTestingPerformancePAR"] =\
        get_average_performance(res_conf, objective)
    test_dict["defaultConfigurationTestingPerformancePAR"] =\
        get_average_performance(res_default, objective)
    smac_run_obj = SMAC2.get_smac_run_obj(
        configurator.scenario.sparkle_objective)
    test_dict["figure-configured-vs-default-test"] =\
        get_figure_configured_vs_default_on_instance_set(
        solver, test_set.name, res_default, res_conf, target_dir, smac_run_obj,
        float(run_cutoff_time),
        configurator.scenario.sparkle_objective, data_type="test")

    # Retrieve timeout numbers for the testing instances
    configured_timeouts_test, default_timeouts_test, overlapping_timeouts_test =\
        get_timeouts_instanceset(solver,
                                 test_set,
                                 configurator,
                                 validator,
                                 run_cutoff_time)

    test_dict["timeoutsTestDefault"] = default_timeouts_test
    test_dict["timeoutsTestConfigured"] = configured_timeouts_test
    test_dict["timeoutsTestOverlap"] = overlapping_timeouts_test
    test_dict["ablationBool"] = get_ablation_bool(ablation)
    test_dict["ablationPath"] = get_ablation_table(ablation)
    return test_dict


def generate_report_for_configuration(solver: Solver,
                                      configurator: Configurator,
                                      validator: Validator,
                                      extractor_dir: Path,
                                      target_path: Path,
                                      latex_template_path: Path,
                                      bibliography_path: Path,
                                      train_set: InstanceSet,
                                      extractor_cuttoff: int,
                                      test_set: InstanceSet = None,
                                      ablation: AblationScenario = None) -> None:
    """Generate a report for algorithm configuration.

    Args:
        solver: Object representation of the solver
        configurator: Configurator for the report
        validator: Validator that validated the configurator
        extractor_dir: Path to the extractor used
        target_path: Where the report files will be placed.
        latex_template_path: Path to the template to use for the report
        bibliography_path: The bib corresponding to the latex template
        train_set: Instance set for training
        extractor_cuttoff: Cut off for extractor
        test_set: Instance set for testing
        ablation: Whether or not ablation is used. Defaults to True.
    """
    target_path.mkdir(parents=True, exist_ok=True)
    variables_dict = configuration_report_variables(
        target_path, solver, configurator, validator, extractor_dir, bibliography_path,
        train_set, extractor_cuttoff, test_set,
        ablation)
    stex.generate_report(latex_template_path,
                         "template-Sparkle-for-configuration.tex",
                         target_path,
                         "Sparkle_Report_for_Configuration",
                         variables_dict)



"""This package provides solver support for Sparkle."""
from sparkle.solver.solver import Solver
from sparkle.solver.extractor import Extractor
from sparkle.solver.selector import Selector

from sparkle.solver.validator import Validator
from sparkle.solver.verifier import SATVerifier, SolutionVerifier
# from sparkle.solver.ablation import AblationScenario  # TODO: Remove cyclic dependency



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Helper functions for ablation analysis."""
from __future__ import annotations
import re
import shutil
import decimal
from pathlib import Path

import runrunner as rrr
from runrunner.base import Runner, Run

from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.help import logging as sl

from sparkle.configurator.implementations import SMAC2
from sparkle.platform import CommandName
from sparkle.solver import Solver
from sparkle.instance import InstanceSet


class AblationScenario:
    """Class for ablation analysis."""
    def __init__(self: AblationScenario,
                 solver: Solver,
                 train_set: InstanceSet,
                 test_set: InstanceSet,
                 output_dir: Path,
                 ablation_executable: Path = None,
                 ablation_validation_executable: Path = None,
                 override_dirs: bool = False) -> None:
        """Initialize ablation scenario.

        Args:
            solver: Solver object
            train_set: The training instance
            test_set: The test instance
            output_dir: The output directory
            ablation_executable: (Only for execution) The ablation executable
            ablation_validation_executable: (Only for execution) The validation exec
            override_dirs: Whether to clean the scenario directory if it already exists
        """
        self.ablation_exec = ablation_executable
        self.ablation_validation_exec = ablation_validation_executable
        self.solver = solver
        self.train_set = train_set
        self.test_set = test_set
        self.output_dir = output_dir
        self.scenario_name = f"{self.solver.name}_{self.train_set.name}"
        if self.test_set is not None:
            self.scenario_name += f"_{self.test_set.name}"
        self.scenario_dir = self.output_dir / self.scenario_name
        if override_dirs and self.scenario_dir.exists():
            print("Warning: found existing ablation scenario. This will be removed.")
            shutil.rmtree(self.scenario_dir)

        # Create required scenario directories
        self.tmp_dir = self.scenario_dir / "tmp"
        self.tmp_dir.mkdir(parents=True, exist_ok=True)

        self.validation_dir = self.scenario_dir / "validation"
        self.validation_dir_tmp = self.validation_dir / "tmp"
        self.validation_dir_tmp.mkdir(parents=True, exist_ok=True)
        self.table_file = self.validation_dir / "log" / "ablation-validation-run1234.txt"

    def create_configuration_file(self: AblationScenario) -> None:
        """Create a configuration file for ablation analysis.

        Args:
            solver: Solver object
            instance_train_name: The training instance
            instance_test_name: The test instance

        Returns:
            None
        """
        ablation_scenario_dir = self.scenario_dir
        objective = gv.settings().get_general_sparkle_objectives()[0]
        configurator = gv.settings().get_general_sparkle_configurator()
        _, opt_config_str = configurator.get_optimal_configuration(
            self.solver, self.train_set, objective=objective)

        # We need to check which params are missing and supplement with default values
        pcs = self.solver.get_pcs()
        for p in pcs:
            if p["name"] not in opt_config_str:
                opt_config_str += f" -{p['name']} {p['default']}"

        # Ablation cannot deal with E scientific notation in floats
        ctx = decimal.Context(prec=16)
        for config in opt_config_str.split(" -"):
            _, value = config.strip().split(" ")
            if "e" in value.lower():
                value = value.strip("'")
                float_value = float(value.lower())
                formatted = format(ctx.create_decimal(float_value), "f")
                opt_config_str = opt_config_str.replace(value, formatted)

        smac_run_obj = SMAC2.get_smac_run_obj(objective)
        objective_str = "MEAN10" if smac_run_obj == "RUNTIME" else "MEAN"
        run_cutoff_time = gv.settings().get_general_target_cutoff_time()
        run_cutoff_length = gv.settings().get_configurator_target_cutoff_length()
        concurrent_clis = gv.settings().get_slurm_max_parallel_runs_per_node()
        ablation_racing = gv.settings().get_ablation_racing_flag()
        configurator = gv.settings().get_general_sparkle_configurator()
        pcs_file_path = f"{self.solver.get_pcs_file().absolute()}"  # Get Solver PCS

        # Create config file
        config_file = Path(f"{ablation_scenario_dir}/ablation_config.txt")
        config = (f'algo = "{configurator.configurator_target.absolute()} '
                  f'{self.solver.directory.absolute()} {objective}"\n'
                  f"execdir = {self.tmp_dir.absolute()}\n"
                  "experimentDir = ./\n"
                  f"deterministic = {1 if self.solver.deterministic else 0}\n"
                  f"run_obj = {smac_run_obj}\n"
                  f"overall_obj = {objective_str}\n"
                  f"cutoffTime = {run_cutoff_time}\n"
                  f"cutoff_length = {run_cutoff_length}\n"
                  f"cli-cores = {concurrent_clis}\n"
                  f"useRacing = {ablation_racing}\n"
                  "seed = 1234\n"
                  f"paramfile = {pcs_file_path}\n"
                  "instance_file = instances_train.txt\n"
                  "test_instance_file = instances_test.txt\n"
                  "sourceConfiguration=DEFAULT\n"
                  f'targetConfiguration="{opt_config_str}"')
        config_file.open("w").write(config)
        # Write config to validation directory
        conf_valid = config.replace(f"execdir = {self.tmp_dir.absolute()}\n",
                                    f"execdir = {self.validation_dir_tmp.absolute()}\n")
        (self.validation_dir / config_file.name).open("w").write(conf_valid)

    def create_instance_file(self: AblationScenario, test: bool = False) -> None:
        """Create an instance file for ablation analysis."""
        file_suffix = "_train.txt"
        instance_set = self.train_set
        if test:
            file_suffix = "_test.txt"
            instance_set = self.test_set if self.test_set is not None else self.train_set
        # We give the Ablation script the paths of the instances
        file_instance = self.scenario_dir / f"instances{file_suffix}"
        with file_instance.open("w") as fh:
            for instance in instance_set._instance_paths:
                # We need to unpack the multi instance file paths in quotes
                if isinstance(instance, list):
                    joined_instances = " ".join(
                        [str(file.absolute()) for file in instance])
                    fh.write(f"{joined_instances}\n")
                else:
                    fh.write(f"{instance.absolute()}\n")
        # Copy to validation directory
        shutil.copyfile(file_instance, self.validation_dir / file_instance.name)

    def check_for_ablation(self: AblationScenario) -> bool:
        """Checks if ablation has terminated successfully."""
        if not self.table_file.is_file():
            return False
        # First line in the table file should be "Ablation analysis validation complete."
        table_line = self.table_file.open().readline().strip()
        return table_line == "Ablation analysis validation complete."

    def read_ablation_table(self: AblationScenario) -> list[list[str]]:
        """Read from ablation table of a scenario."""
        if not self.check_for_ablation():
            # No ablation table exists for this solver-instance pair
            return []
        results = [["Round", "Flipped parameter", "Source value", "Target value",
                    "Validation result"]]

        for line in self.table_file.open().readlines():
            # Pre-process lines from the ablation file and add to the results dictionary.
            # Sometimes ablation rounds switch multiple parameters at once.
            # EXAMPLE: 2 EDR, EDRalpha   0, 0.1   1, 0.1013241633106732 486.31691
            # To split the row correctly, we remove the space before the comma separated
            # parameters and add it back.
            # T.S. 30-01-2024: the results object is a nested list not dictionary?
            values = re.sub(r"\s+", " ", line.strip())
            values = re.sub(r", ", ",", values)
            values = [val.replace(",", ", ") for val in values.split(" ")]
            if len(values) == 5:
                results.append(values)
        return results

    def submit_ablation(self: AblationScenario,
                        run_on: Runner = Runner.SLURM) -> list[Run]:
        """Submit an ablation job.

        Args:
            run_on: Determines to which RunRunner queue the job is added

        Returns:
            A  list of Run objects. Empty when running locally.
        """
        # 1. submit the ablation to the runrunner queue
        clis = gv.settings().get_slurm_max_parallel_runs_per_node()
        cmd = f"{self.ablation_exec.absolute()} --optionFile ablation_config.txt"
        srun_options = ["-N1", "-n1", f"-c{clis}"]
        sbatch_options = [f"--cpus-per-task={clis}"] +\
            gv.settings().get_slurm_extra_options(as_args=True)

        run_ablation = rrr.add_to_queue(
            runner=run_on,
            cmd=cmd,
            name=CommandName.RUN_ABLATION,
            base_dir=sl.caller_log_dir,
            path=self.scenario_dir,
            sbatch_options=sbatch_options,
            srun_options=srun_options)

        runs = []
        if run_on == Runner.LOCAL:
            run_ablation.wait()
        runs.append(run_ablation)

        # 2. Run ablation validation run if we have a test set to run on
        if self.test_set is not None:
            # Validation dir should have a copy of all needed files, except for the
            # output of the ablation run, which is stored in ablation-run[seed].txt
            cmd = f"{self.ablation_validation_exec.absolute()} "\
                  "--optionFile ablation_config.txt "\
                  "--ablationLogFile ../log/ablation-run1234.txt"

            run_ablation_validation = rrr.add_to_queue(
                runner=run_on,
                cmd=cmd,
                name=CommandName.RUN_ABLATION_VALIDATION,
                path=self.validation_dir,
                base_dir=sl.caller_log_dir,
                dependencies=run_ablation,
                sbatch_options=sbatch_options,
                srun_options=srun_options)

            if run_on == Runner.LOCAL:
                run_ablation_validation.wait()
            runs.append(run_ablation_validation)

        return runs



import logging

import numpy as np
import pandas as pd

from sklearn.decomposition import PCA

from ConfigSpace.hyperparameters import CategoricalHyperparameter, \
    UniformFloatHyperparameter, UniformIntegerHyperparameter
from ConfigSpace.conditions import EqualsCondition, InCondition
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace import Configuration

from autofolio.aslib_scenario import ASlibScenario

__author__ = "Marius Lindauer"
__license__ = "BSD"


class PCAWrapper(object):

    @staticmethod
    def add_params(cs: ConfigurationSpace):
        '''
            adds parameters to ConfigurationSpace 
        '''
        pca_switch = CategoricalHyperparameter(
            "pca", choices=[True, False], default_value=False)
        n_components = UniformIntegerHyperparameter(
            "pca_n_components", lower=1, upper=20, default_value=7, log=True)
        cs.add_hyperparameter(pca_switch)
        cs.add_hyperparameter(n_components)
        cond = InCondition(
            child=n_components, parent=pca_switch, values=[True])
        cs.add_condition(cond)

    def __init__(self):
        '''
            Constructor
        '''
        self.pca = None
        self.active = False

        self.logger = logging.getLogger("PCA")

    def fit(self, scenario: ASlibScenario, config: Configuration):
        '''
            fit pca object to ASlib scenario data

            Arguments
            ---------
            scenario: data.aslib_scenario.ASlibScenario
                ASlib Scenario with all data in pandas
            config: ConfigSpace.Configuration
                configuration
        '''

        if config.get("pca"):
            n_components = config.get("pca_n_components")
            if n_components < scenario.feature_data.shape[0] or n_components < scenario.feature_data.shape[1]:
                return
            self.pca = PCA(n_components=config.get("pca_n_components"))
            self.pca.fit(scenario.feature_data.values)
            self.active = True

    def transform(self, scenario: ASlibScenario):
        '''
            transform ASLib scenario data

            Arguments
            ---------
            scenario: data.aslib_scenario.ASlibScenario
                ASlib Scenario with all data in pandas

            Returns
            -------
            data.aslib_scenario.ASlibScenario
        '''
        if self.pca:
            self.logger.debug("Applying PCA")
            values = self.pca.transform(
                np.array(scenario.feature_data.values))

            scenario.feature_data = pd.DataFrame(
                data=values, index=scenario.feature_data.index, columns=["f%d" % (i) for i in range(values.shape[1])])

        return scenario

    def fit_transform(self, scenario: ASlibScenario, config: Configuration):
        '''
            fit and transform

            Arguments
            ---------
            scenario: data.aslib_scenario.ASlibScenario
                ASlib Scenario with all data in pandas
            config: ConfigSpace.Configuration
                configuration

            Returns
            -------
            data.aslib_scenario.ASlibScenario
        '''
        self.fit(scenario, config)
        scenario = self.transform(scenario)
        return scenario
    
    def get_attributes(self):
        '''
            returns a list of tuples of (attribute,value) 
            for all learned attributes
            
            Returns
            -------
            list of tuples of (attribute,value) 
        '''
        return ["Dimensions=%s" %(self.pca.n_components)]



"""This package provides types for Sparkle applications."""
import importlib
import inspect
import re
from typing import Callable

from sparkle.types.sparkle_callable import SparkleCallable
from sparkle.types.features import FeatureGroup, FeatureSubgroup, FeatureType
from sparkle.types.status import SolverStatus
from sparkle.types import objective
from sparkle.types.objective import SparkleObjective, UseTime


objective_string_regex = re.compile(r"(?P<name>[\w\-_]+)(:(?P<direction>min|max))?$")
objective_variable_regex = re.compile(r"(-?\d+)$")


def _check_class(candidate: Callable) -> bool:
    """Verify whether a loaded class is a valid objective class."""
    return inspect.isclass(candidate) and issubclass(candidate, SparkleObjective)


def resolve_objective(objective_name: str) -> SparkleObjective:
    """Try to resolve the objective class by (case-sensitive) name.

    convention: objective_name(variable-k)?(:[min|max])?

    Order of resolving:
        class_name of user defined SparkleObjectives
        class_name of sparkle defined SparkleObjectives
        default SparkleObjective with minimization unless specified as max

    Args:
        name: The name of the objective class. Can include parameter value k.

    Returns:
        Instance of the Objective class or None if not found.
    """
    match = objective_string_regex.fullmatch(objective_name)
    if match is None or objective_name == "" or not objective_name[0].isalpha():
        return None

    name = match.group("name")
    minimise = not match.group("direction") == "max"  # .group returns "" if no match

    # Search for optional variable and record split point between name and variable
    name_options = [(name, None), ]  # Options of names to check for
    if m := objective_variable_regex.search(name):
        argument = int(m.group())
        name_options = [(name[:m.start()], argument), ] + name_options  # Prepend

    # First try to resolve the user input classes
    for rname, rarg in name_options:
        try:
            user_module = importlib.import_module("Settings.objective")
            for o_name, o_class in inspect.getmembers(user_module,
                                                      predicate=_check_class):
                if o_name == rname:
                    if rarg is not None:
                        return o_class(rarg)
                    return o_class()
        except Exception:
            pass

    for rname, rarg in name_options:
        # Try to match with specially defined classes
        for o_name, o_class in inspect.getmembers(objective,
                                                  predicate=_check_class):
            if o_name == rname:
                if rarg is not None:
                    return o_class(rarg)
                return o_class()

    # No special objects found. Return objective with full name
    return SparkleObjective(name=objective_name, minimise=minimise)



#!/usr/bin/env python3
"""Sparkle command to configure a solver."""
from __future__ import annotations

import argparse
import sys
import os
from pathlib import Path
from pandas import DataFrame

from runrunner.base import Runner, Run
import runrunner as rrr

from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.help import logging as sl
from sparkle.platform.settings_objects import SettingState
from sparkle.CLI.help.reporting_scenario import Scenario
from sparkle.structures import FeatureDataFrame
from sparkle.platform import CommandName, COMMAND_DEPENDENCIES
from sparkle.configurator.configurator import Configurator
from sparkle.CLI.help.nicknames import resolve_object_name
from sparkle.solver import Solver
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help import argparse_custom as ac
from sparkle.instance import instance_set, InstanceSet


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(
        description="Configure a solver in the Sparkle platform.",
        epilog=("Note that the test instance set is only used if the ``--ablation``"
                " or ``--validation`` flags are given"))
    parser.add_argument(*ac.ConfiguratorArgument.names,
                        **ac.ConfiguratorArgument.kwargs)
    parser.add_argument(*ac.SolverArgument.names,
                        **ac.SolverArgument.kwargs)
    parser.add_argument(*ac.InstanceSetTrainArgument.names,
                        **ac.InstanceSetTrainArgument.kwargs)
    parser.add_argument(*ac.InstanceSetTestArgument.names,
                        **ac.InstanceSetTestArgument.kwargs)
    parser.add_argument(*ac.SparkleObjectiveArgument.names,
                        **ac.SparkleObjectiveArgument.kwargs)
    parser.add_argument(*ac.TargetCutOffTimeConfigurationArgument.names,
                        **ac.TargetCutOffTimeConfigurationArgument.kwargs)
    parser.add_argument(*ac.WallClockTimeArgument.names,
                        **ac.WallClockTimeArgument.kwargs)
    parser.add_argument(*ac.CPUTimeArgument.names,
                        **ac.CPUTimeArgument.kwargs)
    parser.add_argument(*ac.SolverCallsArgument.names,
                        **ac.SolverCallsArgument.kwargs)
    parser.add_argument(*ac.NumberOfRunsConfigurationArgument.names,
                        **ac.NumberOfRunsConfigurationArgument.kwargs)
    parser.add_argument(*ac.SettingsFileArgument.names,
                        **ac.SettingsFileArgument.kwargs)
    parser.add_argument(*ac.UseFeaturesArgument.names,
                        **ac.UseFeaturesArgument.kwargs)
    parser.add_argument(*ac.ValidateArgument.names,
                        **ac.ValidateArgument.kwargs)
    parser.add_argument(*ac.AblationArgument.names,
                        **ac.AblationArgument.kwargs)
    parser.add_argument(*ac.RunOnArgument.names,
                        **ac.RunOnArgument.kwargs)
    return parser


def apply_settings_from_args(args: argparse.Namespace) -> None:
    """Apply command line arguments to settings.

    Args:
        args: Arguments object created by ArgumentParser.
    """
    if args.settings_file is not None:
        gv.settings().read_settings_ini(args.settings_file, SettingState.CMD_LINE)
    if args.objectives is not None:
        gv.settings().set_general_sparkle_objectives(
            args.objectives, SettingState.CMD_LINE)
    if args.target_cutoff_time is not None:
        gv.settings().set_general_target_cutoff_time(
            args.target_cutoff_time, SettingState.CMD_LINE)
    if args.wallclock_time is not None:
        gv.settings().set_config_wallclock_time(
            args.wallclock_time, SettingState.CMD_LINE)
    if args.cpu_time is not None:
        gv.settings().set_config_cpu_time(
            args.cpu_time, SettingState.CMD_LINE)
    if args.solver_calls is not None:
        gv.settings().set_config_solver_calls(
            args.solver_calls, SettingState.CMD_LINE)
    if args.number_of_runs is not None:
        gv.settings().set_config_number_of_runs(
            args.number_of_runs, SettingState.CMD_LINE)
    if args.run_on is not None:
        gv.settings().set_run_on(
            args.run_on.value, SettingState.CMD_LINE)


def run_after(solver: Path,
              train_set: InstanceSet,
              test_set: InstanceSet,
              dependency: list[Run],
              command: CommandName,
              run_on: Runner = Runner.SLURM) -> Run:
    """Add a command to run after configuration to RunRunner queue.

    Args:
      solver: Path (object) to solver.
      train_set: Instances used for training.
      test_set: Instances used for testing.
      dependency: List of job dependencies.
      command: The command to run. Currently supported: Validation and Ablation.
      run_on: Whether the job is executed on Slurm or locally.

    Returns:
      RunRunner Run object regarding the callback
    """
    cmd_file = "validate_configured_vs_default.py"
    if command == CommandName.RUN_ABLATION:
        cmd_file = "run_ablation.py"

    command_line = f"./sparkle/CLI/{cmd_file} --settings-file Settings/latest.ini "\
                   f"--solver {solver.name} --instance-set-train {train_set.directory}"\
                   f" --run-on {run_on}"
    if test_set is not None:
        command_line += f" --instance-set-test {test_set.directory}"

    run = rrr.add_to_queue(
        runner=run_on,
        cmd=command_line,
        name=command,
        dependencies=dependency,
        base_dir=sl.caller_log_dir,
        srun_options=["-N1", "-n1"],
        sbatch_options=gv.settings().get_slurm_extra_options(as_args=True))

    if run_on == Runner.LOCAL:
        print("Waiting for the local calculations to finish.")
        run.wait()
    return run


if __name__ == "__main__":
    # Log command call
    sl.log_command(sys.argv)

    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args()

    apply_settings_from_args(args)

    validate = args.validate
    ablation = args.ablation
    solver = resolve_object_name(
        args.solver,
        gv.file_storage_data_mapping[gv.solver_nickname_list_path],
        gv.settings().DEFAULT_solver_dir, class_name=Solver)
    instance_set_train = resolve_object_name(
        args.instance_set_train,
        gv.file_storage_data_mapping[gv.instances_nickname_path],
        gv.settings().DEFAULT_instance_dir, instance_set)
    instance_set_test = args.instance_set_test
    if instance_set_test is not None:
        instance_set_test = resolve_object_name(
            args.instance_set_test,
            gv.file_storage_data_mapping[gv.instances_nickname_path],
            gv.settings().DEFAULT_instance_dir, instance_set)
    use_features = args.use_features
    run_on = gv.settings().get_run_on()
    if args.configurator is not None:
        gv.settings().set_general_sparkle_configurator(
            value=getattr(Configurator, args.configurator),
            origin=SettingState.CMD_LINE)

    # Check if Solver and instance sets were resolved
    check_for_initialise(COMMAND_DEPENDENCIES[CommandName.CONFIGURE_SOLVER])

    feature_data_df = None
    if use_features:
        feature_data = FeatureDataFrame(gv.settings().DEFAULT_feature_data_path)

        data_dict = {}
        feature_data_df = feature_data.dataframe

        for label, row in feature_data_df.iterrows():
            # os.path.split(os.path.split(label)[0])[1] gives the dir/instance set name
            if os.path.split(os.path.split(label)[0])[1] == instance_set_train.name:
                if row.empty:
                    print("No feature data exists for the given training set, please "
                          "run add_feature_extractor.py, then compute_features.py")
                    sys.exit(-1)

                new_label = (f"../../../instances/{instance_set_train.name}/"
                             + os.path.split(label)[1])
                data_dict[new_label] = row

        feature_data_df = DataFrame.from_dict(data_dict, orient="index",
                                              columns=feature_data_df.columns)

        if feature_data.has_missing_value():
            print("You have unfinished feature computation jobs, please run "
                  "`sparkle compute features`")
            sys.exit(-1)

        for index, column in enumerate(feature_data_df):
            feature_data_df.rename(columns={column: f"Feature{index+1}"}, inplace=True)

    number_of_runs = gv.settings().get_config_number_of_runs()
    solver_calls = gv.settings().get_config_solver_calls()
    cpu_time = gv.settings().get_config_cpu_time()
    wallclock_time = gv.settings().get_config_wallclock_time()
    cutoff_time = gv.settings().get_general_target_cutoff_time()
    cutoff_length = gv.settings().get_configurator_target_cutoff_length()
    sparkle_objectives =\
        gv.settings().get_general_sparkle_objectives()
    configurator = gv.settings().get_general_sparkle_configurator()
    config_scenario = configurator.scenario_class(
        solver, instance_set_train, number_of_runs, solver_calls, cpu_time,
        wallclock_time, cutoff_time, cutoff_length, sparkle_objectives, use_features,
        configurator.configurator_target, feature_data_df)

    sbatch_options = gv.settings().get_slurm_extra_options(as_args=True)
    dependency_job_list = configurator.configure(
        scenario=config_scenario,
        sbatch_options=sbatch_options,
        num_parallel_jobs=gv.settings().get_number_of_jobs_in_parallel(),
        base_dir=sl.caller_log_dir,
        run_on=run_on)

    # Update latest scenario
    gv.latest_scenario().set_config_solver(solver)
    gv.latest_scenario().set_config_instance_set_train(instance_set_train.directory)
    gv.latest_scenario().set_latest_scenario(Scenario.CONFIGURATION)

    if instance_set_test is not None:
        gv.latest_scenario().set_config_instance_set_test(instance_set_test.directory)
    else:
        # Set to default to overwrite possible old path
        gv.latest_scenario().set_config_instance_set_test()

    # Set validation to wait until configuration is done
    if validate:
        validate_jobid = run_after(
            solver, instance_set_train, instance_set_test, dependency_job_list,
            command=CommandName.VALIDATE_CONFIGURED_VS_DEFAULT, run_on=run_on
        )
        dependency_job_list.append(validate_jobid)

    if ablation:
        ablation_jobid = run_after(
            solver, instance_set_train, instance_set_test, dependency_job_list,
            command=CommandName.RUN_ABLATION, run_on=run_on
        )
        dependency_job_list.append(ablation_jobid)

    if run_on == Runner.SLURM:
        job_id_str = ",".join([run.run_id for run in dependency_job_list])
        print(f"Running configuration. Waiting for Slurm job(s) with id(s): "
              f"{job_id_str}")
    else:
        print("Running configuration finished!")

    # Write used settings to file
    gv.settings().write_used_settings()
    # Write used scenario to file
    gv.latest_scenario().write_scenario_ini()
